# (C) 2025 GoodData Corporation

from typing import Any, Callable

from gooddata_pipelines.logger.logger import LogObserver

logger: LogObserver = LogObserver()


def log_and_reraise_exception(message: str) -> Callable:
    """
    Decorator to log an exception and re-raise it.

    Args:
        message (str): The message to log.
    """

    def decorator(fn: Callable) -> Callable:
        def wrapper(*method_args: Any, **method_kwargs: Any) -> Callable:
            try:
                return fn(*method_args, **method_kwargs)
            except Exception:
                logger.error(
                    f"{message}, {fn.__name__}, Args: {method_args}, Kwargs: {method_kwargs}"
                )
                raise

        return wrapper

    return decorator



# (C) 2025 GoodData Corporation

import os
import tempfile

import pytest

from gooddata_pipelines.backup_and_restore.backup_input_processor import (
    BackupInputProcessor,
)
from gooddata_pipelines.backup_and_restore.models.input_type import InputType
from gooddata_pipelines.backup_and_restore.models.workspace_response import (
    Hierarchy,
    Links,
    Meta,
    Page,
    Workspace,
    WorkspaceResponse,
)


@pytest.fixture
def backup_input_processor(mock_gooddata_api, mock_logger):
    processor = BackupInputProcessor(mock_gooddata_api, page_size=2)
    processor.hierarchy_endpoint = (
        "/fake/hierarchy?filter=parent.id=={parent_id}"
    )
    processor.all_workspaces_endpoint = "/fake/all"
    processor.logger.subscribe(mock_logger)
    return processor


def test_process_data_extracts_children_and_subparents(backup_input_processor):
    ws1 = Workspace(id="ws1", meta=Meta(hierarchy=Hierarchy(children_count=2)))
    ws2 = Workspace(id="ws2", meta=Meta(hierarchy=Hierarchy(children_count=0)))
    ws3 = Workspace(id="ws3", meta=None)

    result = backup_input_processor.process_data([ws1, ws2, ws3])
    assert result.workspace_ids == ["ws1", "ws2", "ws3"]
    assert result.sub_parents == ["ws1"]


def test_log_paging_progress_logs_info(backup_input_processor, capsys):
    response = WorkspaceResponse(
        data=[],
        meta=Meta(
            page=Page(size=5, total_elements=25, number=1, total_pages=5),
            hierarchy=None,
        ),
        links=Links(self="self", next="next"),
    )

    backup_input_processor.log_paging_progress(response)
    captured = capsys.readouterr()
    assert "Fetched page: 2 of 5" in captured.out


def test_log_paging_progress_no_page(backup_input_processor, capsys):
    response = WorkspaceResponse(
        data=[],
        meta=Meta(page=None, hierarchy=None),
        links=Links(self="self", next="next"),
    )

    backup_input_processor.log_paging_progress(response)
    captured = capsys.readouterr()
    assert captured.out == ""


def test_paginate_calls_fetch_page_and_process_data(
    backup_input_processor, monkeypatch
):
    ws1 = Workspace(id="ws1", meta=Meta(hierarchy=Hierarchy(children_count=1)))
    ws2 = Workspace(id="ws2", meta=Meta(hierarchy=Hierarchy(children_count=0)))
    links1 = Links(self="self", next="next_url")
    links2 = Links(self="self", next=None)
    resp1 = WorkspaceResponse(
        data=[ws1], meta=Meta(hierarchy=None, page=None), links=links1
    )
    resp2 = WorkspaceResponse(
        data=[ws2], meta=Meta(hierarchy=None, page=None), links=links2
    )

    fetch_page_calls = []

    def fetch_page_side_effect(url):
        fetch_page_calls.append(url)
        return resp1 if len(fetch_page_calls) == 1 else resp2

    backup_input_processor.fetch_page = fetch_page_side_effect

    process_data_calls = []

    def process_data_side_effect(data):
        process_data_calls.append(data)
        if len(process_data_calls) == 1:
            return BackupInputProcessor._ProcessDataOutput(["ws1"], ["ws1"])
        else:
            return BackupInputProcessor._ProcessDataOutput(["ws2"], [])

    monkeypatch.setattr(
        BackupInputProcessor,
        "process_data",
        staticmethod(process_data_side_effect),
    )
    monkeypatch.setattr(
        BackupInputProcessor,
        "log_paging_progress",
        staticmethod(lambda resp: None),
    )

    result = backup_input_processor._paginate("first_url")
    assert len(result) == 2
    assert result[0].workspace_ids == ["ws1"]
    assert result[1].workspace_ids == ["ws2"]
    assert len(fetch_page_calls) == 2
    assert len(process_data_calls) == 2


def test_get_hierarchy_recurses(backup_input_processor):
    def fake_paginate(url):
        if "p1" in url:
            return [BackupInputProcessor._ProcessDataOutput(["c1"], ["c1"])]
        if "c1" in url:
            return [BackupInputProcessor._ProcessDataOutput(["c2"], [])]
        return []

    backup_input_processor._paginate = fake_paginate
    result = backup_input_processor.get_hierarchy("p1")
    assert set(result) == {"c1", "c2"}


def test_get_workspaces_to_backup_empty_org(
    backup_input_processor, monkeypatch, capsys
):
    """Test that the function returns an empty list if the organization contains no workspaces."""
    monkeypatch.setattr(
        backup_input_processor,
        "_paginate",
        lambda _: [],
    )
    backup_input_processor.get_ids_to_backup(
        InputType.ORGANIZATION,
        "some-csv-file.csv",
    )
    captured = capsys.readouterr()
    assert "No workspaces found in the organization." in captured.out


def test_read_csv_input_empty_file(backup_input_processor) -> None:
    """Test with an empty CSV file."""
    with tempfile.NamedTemporaryFile() as temp_csv:
        path_to_csv = temp_csv.name
        with pytest.raises(
            ValueError, match="No content found in the CSV file."
        ):
            backup_input_processor.csv_reader.read_backup_csv(path_to_csv)


def test_read_csv_input_only_header(backup_input_processor) -> None:
    """Test with a CSV file that contains only the header."""
    with tempfile.NamedTemporaryFile() as temp_csv:
        temp_csv.write(b"header1\n")
        temp_csv.flush()
        temp_csv.seek(0)
        path_to_csv = temp_csv.name
        with pytest.raises(
            ValueError, match="No workspaces found in the CSV file."
        ):
            backup_input_processor.csv_reader.read_backup_csv(path_to_csv)


def test_read_csv_input_valid(backup_input_processor) -> None:
    """Test with a valid CSV file."""
    with tempfile.NamedTemporaryFile(delete=False) as temp_csv:
        temp_csv.write(b"header1\n")
        temp_csv.write(b"workspace1\n")
        temp_csv.write(b"workspace2\n")
        temp_csv.flush()
        temp_csv.seek(0)
        path_to_csv = temp_csv.name
        result = backup_input_processor.csv_reader.read_backup_csv(path_to_csv)
        assert result == ["workspace1", "workspace2"]
    os.remove(path_to_csv)


def test_read_csv_input_too_many_columns(backup_input_processor) -> None:
    """Test with a CSV file that contains too many columns."""
    with tempfile.NamedTemporaryFile(delete=False) as temp_csv:
        temp_csv.write(b"header1,header2\n")
        temp_csv.write(b"workspace1,extra_column\n")
        temp_csv.flush()
        temp_csv.seek(0)
        path_to_csv = temp_csv.name
        with pytest.raises(
            ValueError,
            match="Input file contains more than one column. Please check the input and try again.",
        ):
            backup_input_processor.csv_reader.read_backup_csv(path_to_csv)
    os.remove(path_to_csv)



# (C) 2025 GoodData Corporation

import abc

from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
)
from gooddata_pipelines.logger import LogObserver


class BackupStorage(abc.ABC):
    def __init__(self, conf: BackupRestoreConfig):
        self.logger = LogObserver()

        suffix = "/" if not conf.storage.backup_path.endswith("/") else ""
        self._backup_path = conf.storage.backup_path + suffix

    @abc.abstractmethod
    def export(self, folder: str, org_id: str) -> None:
        """Exports the content of the folder to the storage."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_ws_declaration(
        self, target_path: str, local_target_path: str
    ) -> None:
        raise NotImplementedError(
            "This method should be implemented by the subclass."
        )



# (C) 2025 GoodData Corporation
from typing import Literal, Optional

import attrs
import orjson
import pytest
from gooddata_api_client.exceptions import NotFoundException  # type: ignore
from gooddata_sdk.catalog.user.entity_model.user import (
    CatalogUser,
    CatalogUserAttributes,
    CatalogUserGroupsData,
    CatalogUserRelationships,
)
from gooddata_sdk.catalog.user.entity_model.user_group import (
    CatalogUserGroup,
)
from pytest_mock import MockerFixture
from requests import Response

from gooddata_pipelines.provisioning.entities.users.models.users import (
    UserFullLoad,
    UserIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.users.users import (
    UserProvisioner,
)
from tests.conftest import TEST_DATA_DIR

TEST_DATA_SUBDIR = f"{TEST_DATA_DIR}/provisioning/entities/users"


@attrs.define
class MockUser:
    id: str
    firstname: Optional[str]
    lastname: Optional[str]
    email: Optional[str]
    authenticationId: Optional[str]
    user_groups: list[str]

    def to_sdk(self):
        return CatalogUser.init(
            user_id=self.id,
            firstname=self.firstname,
            lastname=self.lastname,
            email=self.email,
            authentication_id=self.authenticationId,
            user_group_ids=self.user_groups,
        )

    def to_json(self):
        attrs = {}
        if self.authenticationId:
            attrs["authenticationId"] = self.authenticationId
        if self.firstname:
            attrs["firstname"] = self.firstname
        if self.lastname:
            attrs["lastname"] = self.lastname
        if self.email:
            attrs["email"] = self.email

        data = {
            "id": self.id,
            "type": "user",
            "attributes": attrs,
        }

        if not self.user_groups:
            return data

        relsdata = [
            {"id": group, "type": "userGroup"} for group in self.user_groups
        ]
        if relsdata:
            data["relationships"] = {"userGroups": {"data": relsdata}}
        return data


def test_user_obj_from_sdk():
    user_input = MockUser(
        "some.user", "some", "user", "some@email.com", "auth", ["ug"]
    )
    excepted = UserIncrementalLoad(
        user_id="some.user",
        firstname="some",
        lastname="user",
        email="some@email.com",
        auth_id="auth",
        user_groups=["ug"],
        is_active=True,
    )
    user = UserIncrementalLoad.from_sdk_obj(user_input.to_sdk())
    assert excepted == user


def test_user_obj_from_sdk_no_ugs():
    user_input = MockUser(
        "some.user", "some", "user", "some@email.com", "auth", []
    )
    excepted = UserIncrementalLoad(
        user_id="some.user",
        firstname="some",
        lastname="user",
        email="some@email.com",
        auth_id="auth",
        user_groups=[],
        is_active=True,
    )
    user = UserIncrementalLoad.from_sdk_obj(user_input.to_sdk())
    assert excepted == user


def test_user_obj_to_sdk():
    user_input = MockUser(
        "some.user", "some", "user", "some@email.com", "auth", ["ug"]
    )
    user = UserIncrementalLoad(
        user_id="some.user",
        firstname="some",
        lastname="user",
        email="some@email.com",
        auth_id="auth",
        user_groups=["ug"],
        is_active=True,
    )
    excepted = user_input.to_sdk()
    assert excepted == user.to_sdk_obj()


def test_user_obj_to_sdk_no_ugs():
    user_input = MockUser(
        "some.user", "some", "user", "some@email.com", "auth", []
    )
    user = UserIncrementalLoad(
        user_id="some.user",
        firstname="some",
        lastname="user",
        email="some@email.com",
        auth_id="auth",
        user_groups=[],
        is_active=True,
    )
    excepted = user_input.to_sdk()
    assert excepted == user.to_sdk_obj()


@pytest.fixture
def user_provisioner(mocker: MockerFixture) -> UserProvisioner:
    """Mock instance of UserProvisioner."""
    provisioner_instance = UserProvisioner.create(
        host="https://localhost:3000", token="token"
    )

    # Patch the API
    mocker.patch.object(provisioner_instance, "_api", return_value=None)

    return provisioner_instance


def parse_user_data(user_data: list[dict]) -> list[CatalogUser]:
    """Parse json user metadata to CatalogUser objects."""
    users: list[CatalogUser] = []
    for user in user_data:
        users.append(
            CatalogUser(
                id=user["user_id"],
                attributes=CatalogUserAttributes(
                    firstname=user["firstname"],
                    lastname=user["lastname"],
                    email=user["email"],
                    authentication_id=user["authentication_id"],
                ),
                relationships=CatalogUserRelationships(
                    user_groups=CatalogUserGroupsData(
                        data=[
                            CatalogUserGroup(id=group)
                            for group in user["user_groups"]
                        ]
                    )
                ),
            )
        )
    return sorted(users, key=lambda x: x.id)


@pytest.mark.parametrize(
    ("input_path", "expected_path", "load_method"),
    [
        (
            "users_input_full_load.json",
            "users_expected_full_load.json",
            "full_load",
        ),
        (
            "users_input_incremental_load.json",
            "users_expected_incremental_load.json",
            "incremental_load",
        ),
        (
            "users_input_full_load_modifies_protected_user.json",
            "users_expected_full_load.json",
            "full_load",
        ),
        (
            "users_input_incremental_load_modifies_protected_user.json",
            "users_expected_incremental_load.json",
            "incremental_load",
        ),
        (
            "users_input_incremental_load_deletes_protected_user.json",
            "users_expected_incremental_load.json",
            "incremental_load",
        ),
    ],
)
def test_user_provisioning(
    input_path: str,
    expected_path: str,
    load_method: Literal["full_load", "incremental_load"],
    user_provisioner: UserProvisioner,
    mocker: MockerFixture,
):
    """Test complete user provisioning workflow by checking that the script will
    attempt to create, update or delete expected users for given input."""

    # Load input data
    with open(f"{TEST_DATA_SUBDIR}/{input_path}", "r") as f:
        input_data = orjson.loads(f.read())

    # Load expected data
    with open(f"{TEST_DATA_SUBDIR}/{expected_path}", "r") as f:
        raw_expected_data = orjson.loads(f.read())

    # Load and patch "existing users"
    with open(f"{TEST_DATA_SUBDIR}/existing_upstream_users.json", "r") as f:
        raw_upstream_users = orjson.loads(f.read())

    upstream_users = parse_user_data(raw_upstream_users)

    mocker.patch.object(
        user_provisioner._api,
        "list_users",
        return_value=upstream_users,
    )

    def mock_get_profile(*args, **kwargs) -> Response:
        """Mock the get_profile method by creating a response object with sample
        response from the API reference.
        """
        with open(
            f"{TEST_DATA_SUBDIR}/profile_response_content.json", "r"
        ) as f:
            profile_response = f.read()
        response = Response()
        response.status_code = 200
        response._content = profile_response.encode("utf-8")
        response.headers["Content-Type"] = "application/json"
        return response

    mocker.patch.object(
        user_provisioner._api,
        "get_profile",
        side_effect=mock_get_profile,
    )

    upstream_user_cache = {user.id: user for user in upstream_users}

    def patch_get_user(user_id: str):
        if user_id in upstream_user_cache:
            return upstream_user_cache[user_id]
        raise NotFoundException(f"User {user_id} not found")

    mocker.patch.object(
        user_provisioner._api._sdk.catalog_user,
        "get_user",
        side_effect=patch_get_user,
    )

    # Parse expected data
    expected_deleted_users = sorted(raw_expected_data["deleted_users"])
    raw_expected_modified_users = raw_expected_data["modified_users"]

    expected_modified_users = parse_user_data(raw_expected_modified_users)

    # Patch the API methods to store which users were modified or deleted
    created_or_updated_users: list[CatalogUser] = []
    deleted_users: list[str] = []

    def patch_create_or_update_user(user: CatalogUser, *args, **kwargs):
        created_or_updated_users.append(user)

    def patch_delete_user(user_id: str, *args, **kwargs):
        deleted_users.append(user_id)

    mocker.patch.object(
        user_provisioner._api,
        "create_or_update_user",
        side_effect=patch_create_or_update_user,
    )
    mocker.patch.object(
        user_provisioner._api,
        "delete_user",
        side_effect=patch_delete_user,
    )

    # Run the provisioning
    if load_method == "incremental_load":
        incremental_load_data = [
            UserIncrementalLoad(**row) for row in input_data
        ]
        user_provisioner.incremental_load(incremental_load_data)
    else:
        full_load_data = [UserFullLoad(**row) for row in input_data]
        user_provisioner.full_load(full_load_data)

    # Compare list lengths
    assert len(created_or_updated_users) == len(expected_modified_users)
    assert len(deleted_users) == len(expected_deleted_users)

    # Sort the actual data
    created_or_updated_users = sorted(
        created_or_updated_users, key=lambda x: x.id
    )
    deleted_users = sorted(deleted_users)

    # Compare the actual data
    assert deleted_users == expected_deleted_users
    assert created_or_updated_users == expected_modified_users



# (C) 2025 GoodData Corporation

"""Exception class for Panther operations.

This module defines the internally used `PantherException` class, which is used
to handle exceptions that occur during operations related to the Panther SDK or
GoodData Cloud API.
"""


class GoodDataApiException(Exception):
    """Exception raised during Panther operations.

    This exception is used to indicate errors that occur during operations
    related to interactions with the GoodData Python SDK or GoodData Cloud API.
    It can include additional context provided through keyword arguments.
    """

    def __init__(self, message: str, **kwargs: str) -> None:
        """Raise a PantherException with a message and optional context.

        Args:
            message (str): The error message describing the exception.
            **kwargs: Additional context for the exception, such as HTTP status,
                    API endpoint, and HTTP method or any other relevant information.
        """

        super().__init__(message)
        self.error_message: str = message

        # Set default values for attributes.
        # TODO: Consider if the defaults for these are still needed
        # - the values were necessary for log schema implementations, which
        #   are not used anymore.
        self.http_status: str = "500 Internal Server Error"
        self.api_endpoint: str = "NA"
        self.http_method: str = "NA"

        # Set attributes from kwargs.
        for key, value in kwargs.items():
            setattr(self, key, value)



# (C) 2025 GoodData Corporation
"""Module for provisioning workspaces in GoodData Cloud."""

from typing import Literal

from gooddata_sdk.catalog.workspace.entity_model.workspace import (
    CatalogWorkspace,
)

from gooddata_pipelines.api.exceptions import GoodDataApiException
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceDataMaps,
    WorkspaceFullLoad,
    WorkspaceIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_filters import (
    WorkspaceDataFilterManager,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_parser import (
    WorkspaceDataParser,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_validator import (
    WorkspaceDataValidator,
)
from gooddata_pipelines.provisioning.provisioning import Provisioning
from gooddata_pipelines.provisioning.utils.context_objects import (
    WorkspaceContext,
)
from gooddata_pipelines.provisioning.utils.exceptions import WorkspaceException


class WorkspaceProvisioner(
    Provisioning[WorkspaceFullLoad, WorkspaceIncrementalLoad]
):
    source_group_full: list[WorkspaceFullLoad]
    source_group_incremental: list[WorkspaceIncrementalLoad]

    FULL_LOAD_TYPE: type[WorkspaceFullLoad] = WorkspaceFullLoad
    INCREMENTAL_LOAD_TYPE: type[WorkspaceIncrementalLoad] = (
        WorkspaceIncrementalLoad
    )

    upstream_group: list[CatalogWorkspace]

    def __init__(self, *args: str, **kwargs: str) -> None:
        """Creates an instance of the WorkspaceProvisioner.

        Calls the superclass constructor and initializes the validator, parser,
        and maps for workspace data. Use the `full_load` or `incremental_load`
        methods to run the provisioning.
        """
        super().__init__(*args, **kwargs)
        self.validator: WorkspaceDataValidator = WorkspaceDataValidator(
            self._api
        )
        self.parser: WorkspaceDataParser = WorkspaceDataParser()
        self.maps: WorkspaceDataMaps = WorkspaceDataMaps()

    def _find_workspaces_to_update(
        self,
        source_group: list[WorkspaceFullLoad],
        panther_group: list[CatalogWorkspace],
        ids_in_both_systems: set[str],
    ) -> set[str]:
        """
        Inspects existing Panther workspaces and compares them to workspaces from
        the source database. If the ID exists in both systems but the workspace
        name in GoodData Cloud is different from the source, the workspace will
        be updated. The rest of the workspaces will be ignored.
        """
        existing_workspaces: dict[str, CatalogWorkspace] = {
            workspace.id: workspace for workspace in panther_group
        }

        ids_to_update: set[str] = set()

        for source_workspace in source_group:
            source_id = source_workspace.workspace_id
            source_name = source_workspace.workspace_name

            if source_id not in ids_in_both_systems:
                continue

            if existing_workspaces.get(source_id):
                panther_name = existing_workspaces[source_id].name
            else:
                continue

            if source_name == panther_name:
                continue

            ids_to_update.add(source_id)

        return ids_to_update

    def _create_or_update_panther_workspaces(
        self,
        workspace_ids_to_create: set[str],
        workspace_ids_to_update: set[str],
        child_to_parent_map: dict[str, str],
        workspace_id_to_wdf_map: dict[str, dict[str, list[str]]],
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
    ) -> None:
        action: Literal["CREATE", "UPDATE"]

        for source_workspace in source_group:
            if source_workspace.workspace_id in workspace_ids_to_update:
                action = "UPDATE"
            elif source_workspace.workspace_id in workspace_ids_to_create:
                action = "CREATE"
            else:
                continue

            context: WorkspaceContext = WorkspaceContext(
                workspace_id=source_workspace.workspace_id,
                workspace_name=source_workspace.workspace_name,
                wdf_id=source_workspace.workspace_data_filter_id,
                wdf_values=source_workspace.workspace_data_filter_values,
            )

            parent_workspace_id: str = child_to_parent_map[context.workspace_id]

            try:
                self._api.create_or_update_panther_workspace(
                    workspace_id=context.workspace_id,
                    workspace_name=str(context.workspace_name),
                    parent_id=parent_workspace_id,
                )
                self.logger.info(
                    f"{action.title()}d workspace: {context.workspace_id}"
                )

            except GoodDataApiException as e:
                combined_context = {**context.__dict__, **e.__dict__}
                self.logger.error(
                    f"Failed to {action.title()} workspace: {context.workspace_id}. "
                    + f"Error: {e} Context: {combined_context}"
                )

            # If child workspace has WDF settings, apply them
            child_wdfs: dict[str, list[str]] = workspace_id_to_wdf_map.get(
                context.workspace_id, {}
            )
            if child_wdfs:
                self.wdf_manager.check_wdf_settings(
                    context,
                )

    def delete_panther_workspaces(
        self, ids_to_delete: set[str], workspace_id_to_name_map: dict[str, str]
    ) -> None:
        for workspace_id in ids_to_delete:
            workspace_context: WorkspaceContext = WorkspaceContext(
                workspace_id=workspace_id,
                workspace_name=workspace_id_to_name_map.get(workspace_id),
            )
            try:
                self._api.delete_panther_workspace(workspace_id)
                self.logger.info(
                    f"Deleted workspace: {workspace_context.workspace_id}"
                )

            except GoodDataApiException as e:
                exception_context = {**workspace_context.__dict__, **e.__dict__}
                self.logger.error(
                    f"Failed to delete workspace: {workspace_context.workspace_id}. "
                    + f"Error: {e} Context: {exception_context}"
                )

    def verify_workspace_provisioning(
        self,
        source_group: list[WorkspaceFullLoad],
        parent_workspace_ids: set[str],
    ) -> None:
        """Verifies that upstream content is equal to the source data."""
        source_ids_names: set[tuple[str, str]] = {
            (item.workspace_id, item.workspace_name) for item in source_group
        }

        panther_workspaces: list[CatalogWorkspace] = (
            self._api.get_panther_children_workspaces(parent_workspace_ids)
        )

        panther_ids_names: set[tuple[str, str]] = {
            (workspace.workspace_id, workspace.name)
            for workspace in panther_workspaces
        }

        diff: set[tuple[str, str]] = source_ids_names.symmetric_difference(
            panther_ids_names
        )

        if diff:
            raise WorkspaceException(
                "Provisioning failed. The source and Panther workspaces do not "
                + f"match. Difference: {diff}"
            )

    def _provision_full_load(self) -> None:
        """Full load workspace provisioning."""

        # Validate the source data.
        self.validator.validate_source_data(self.source_group_full)

        # Set the maps based on the source data.
        self.maps = self.parser.set_maps_based_on_source(
            self.maps, self.source_group_full
        )

        # Get upstream children of all parent workspaces.
        self.upstream_group = self._api.get_panther_children_workspaces(
            self.maps.parent_ids
        )

        # Set maps that require upstream data.
        self.maps = self.parser.set_maps_with_upstream_data(
            self.maps, self.source_group_full, self.upstream_group
        )

        # Create an instance of WDF manager with the created maps.
        self.wdf_manager = WorkspaceDataFilterManager(self._api, self.maps)

        # Sort the ids to groups based on provisioning logic.
        id_groups = self._create_groups(
            self.maps.source_ids, self.maps.upstream_ids
        )

        # Find out which workspaces to update.
        self.ids_to_update: set[str] = self._find_workspaces_to_update(
            self.source_group_full,
            self.upstream_group,
            id_groups.ids_in_both_systems,
        )

        # Delete the workspaces that are not in the source.
        self.delete_panther_workspaces(
            id_groups.ids_to_delete, self.maps.workspace_id_to_name_map
        )

        # Create or update selected workspaces.
        self._create_or_update_panther_workspaces(
            id_groups.ids_to_create,
            self.ids_to_update,
            self.maps.child_to_parent_id_map,
            self.maps.workspace_id_to_wdf_map,
            self.source_group_full,
        )

        # Check WDF settings of ignored workspaces.
        ignored_workspace_ids: set[str] = self.maps.source_ids.difference(
            id_groups.ids_to_create.union(self.ids_to_update).union(
                id_groups.ids_to_delete
            )
        )

        for ignored_workspace_id in ignored_workspace_ids:
            ignored_workspace_context: WorkspaceContext = WorkspaceContext(
                workspace_id=ignored_workspace_id,
                workspace_name=self.maps.workspace_id_to_name_map.get(
                    ignored_workspace_id
                ),
            )
            self.wdf_manager.check_wdf_settings(ignored_workspace_context)

        # Verify the provisioning by queries to GoodData Cloud.
        self.verify_workspace_provisioning(
            self.source_group_full, self.maps.parent_ids
        )

    def _provision_incremental_load(self) -> None:
        """Incremental workspace provisioning."""
        # Set the maps based on the source data.
        self.maps = self.parser.set_maps_based_on_source(
            self.maps, self.source_group_incremental
        )

        # Get upstream children of all parent workspaces.
        self.upstream_group = self._api.get_panther_children_workspaces(
            self.maps.parent_ids
        )

        # Set maps that require upstream data.
        self.maps = self.parser.set_maps_with_upstream_data(
            self.maps, self.source_group_incremental, self.upstream_group
        )

        # Create an instance of WDF manager with the created maps.
        self.wdf_manager = WorkspaceDataFilterManager(self._api, self.maps)

        # Iterate through the source data and sort workspace ID to groups
        ids_to_update: set[str] = set()
        ids_to_delete: set[str] = set()

        for workspace in self.source_group_incremental:
            if workspace.is_active:
                ids_to_update.add(workspace.workspace_id)
            else:
                ids_to_delete.add(workspace.workspace_id)

        self._create_or_update_panther_workspaces(
            set(),
            ids_to_update,
            self.maps.child_to_parent_id_map,
            self.maps.workspace_id_to_wdf_map,
            self.source_group_incremental,
        )

        self.delete_panther_workspaces(
            ids_to_delete, self.maps.workspace_id_to_name_map
        )



# (C) 2025 GoodData Corporation

import pytest
from gooddata_api_client import ApiException  # type: ignore[import]
from gooddata_sdk.catalog.workspace.entity_model.workspace import (
    CatalogWorkspace,
)

from gooddata_pipelines.api.exceptions import GoodDataApiException
from gooddata_pipelines.api.utils import raise_with_context

GOODDATA_WRAPPER_OBJECT_PATH = (
    "gooddata_pipelines.api.gooddata_api_wrapper.GoodDataApi.list_workspaces"
)


def test_raise_with_context_reraises_panther_exception():
    @raise_with_context()
    def func():
        raise GoodDataApiException("Caught a GoodDataApiException")

    with pytest.raises(GoodDataApiException) as exc:
        func()
    assert "GoodDataApiException: Caught a GoodDataApiException" in str(
        exc.value
    )


def test_raise_with_context_wraps_generic_exception():
    @raise_with_context(api_endpoint="some.function.name")
    def func():
        raise ValueError("fail")

    with pytest.raises(GoodDataApiException) as exc:
        func()
        assert "ValueError: Some error: fail" in str(exc.value)
        assert exc.value.api_endpoint == "some.function.name"


def test_raise_with_context_wraps_apiexception_and_sets_error_template():
    @raise_with_context(api_endpoint="endpoint")
    def func():
        raise ApiException(404, "Not Found")

    with pytest.raises(GoodDataApiException) as exc:
        func()
        assert "ApiException: 404 Not Found" in str(exc.value)
        assert exc.value.api_endpoint == "endpoint"
        assert exc.value.http_status == "404 Not Found"


def test_raise_with_context_passes_method_kwargs():
    @raise_with_context()
    def func(**kwargs):
        raise RuntimeError("fail")

    with pytest.raises(GoodDataApiException) as exc:
        func(http_method="bar")
        assert exc.value.http_method == "bar"


def test_get_panther_children_workspaces_empty_response(
    mock_gooddata_api, mocker
) -> None:
    parent_ids: set[str] = {"parent_id_1", "parent_id_2"}

    mocker.patch(
        GOODDATA_WRAPPER_OBJECT_PATH,
        return_value=[],
    )

    panther_children = mock_gooddata_api.get_panther_children_workspaces(
        parent_ids
    )

    assert panther_children == []


def test_get_panther_children_full_match(mock_gooddata_api, mocker) -> None:
    parent_ids: set[str] = {"parent_id_1", "parent_id_2"}

    mocker.patch(
        GOODDATA_WRAPPER_OBJECT_PATH,
        return_value=[
            CatalogWorkspace(
                workspace_id="workspace_id1",
                name="workspace_title1",
                parent_id="parent_id_1",
            ),
            CatalogWorkspace(
                workspace_id="workspace_id2",
                name="workspace_title2",
                parent_id="parent_id_2",
            ),
        ],
    )

    panther_children = mock_gooddata_api.get_panther_children_workspaces(
        parent_ids
    )

    assert len(panther_children) == 2
    assert panther_children[0].workspace_id == "workspace_id1"
    assert panther_children[1].workspace_id == "workspace_id2"


def test_get_panther_children_no_match(mock_gooddata_api, mocker) -> None:
    parent_ids: set[str] = {"parent_id_3", "parent_id_4"}

    mocker.patch(
        GOODDATA_WRAPPER_OBJECT_PATH,
        return_value=[
            CatalogWorkspace(
                workspace_id="workspace_id1",
                name="workspace_title1",
                parent_id="parent_id_1",
            ),
            CatalogWorkspace(
                workspace_id="workspace_id2",
                name="workspace_title2",
                parent_id="parent_id_2",
            ),
        ],
    )

    panther_children = mock_gooddata_api.get_panther_children_workspaces(
        parent_ids
    )

    assert len(panther_children) == 0



# (C) 2025 GoodData Corporation

import shutil
from pathlib import Path

from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
    LocalStorageConfig,
)
from gooddata_pipelines.backup_and_restore.storage.base_storage import (
    BackupStorage,
)


class LocalStorage(BackupStorage):
    def __init__(self, conf: BackupRestoreConfig):
        super().__init__(conf)
        if not isinstance(conf.storage, LocalStorageConfig):
            raise ValueError("Local storage config is required")
        self._config: LocalStorageConfig = conf.storage

    def _export(self, folder: str, org_id: str, export_folder: str) -> None:
        """Copies the content of the folder to local storage as backup."""

        self.logger.info(f"Saving {org_id} to local storage")
        shutil.copytree(
            Path(folder), Path(Path.cwd(), export_folder), dirs_exist_ok=True
        )

    def export(self, folder: str, org_id: str) -> None:
        """Copies the content of the folder to local storage as backup."""
        try:
            self._export(folder, org_id, self._config.backup_path)
        except Exception as e:
            self.logger.error(
                f"Error exporting {folder} to {self._config.backup_path}: {e}"
            )
            raise

    def get_ws_declaration(
        self, target_path: str, local_target_path: str
    ) -> None:
        """Retrieves workspace declaration from local storage and copies to the local target path.

        The local target should be a temporary directory.
        """
        file_to_copy = self._backup_path + target_path + "/gooddata_layouts.zip"
        shutil.copy(file_to_copy, local_target_path)



# (C) 2025 GoodData Corporation

"""Module for parsing and processing workspace data in GoodData Cloud."""

from gooddata_sdk.catalog.workspace.entity_model.workspace import (
    CatalogWorkspace,
)

from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceDataMaps,
    WorkspaceFullLoad,
    WorkspaceIncrementalLoad,
)


class WorkspaceDataParser:
    """Helper class to process workspace data retrieved from Panther and source DB."""

    @staticmethod
    def _get_id_to_name_map(
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
        upstream_group: list[CatalogWorkspace],
    ) -> dict[str, str]:
        """Creates a map of workspace IDs to their names for all known workspaces."""
        source_map: dict[str, str] = {
            workspace.workspace_id: workspace.workspace_name
            for workspace in source_group
        }
        upstream_map: dict[str, str] = {
            item.workspace_id: item.name for item in upstream_group
        }

        return {**upstream_map, **source_map}

    @staticmethod
    def _get_child_to_parent_map(
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
    ) -> dict[str, str]:
        """Creates a map of child workspace IDs to their parent workspace IDs."""
        child_to_parent_map: dict[str, str] = {
            workspace.workspace_id: workspace.parent_id
            for workspace in source_group
        }

        return child_to_parent_map

    @staticmethod
    def _get_set_of_ids_from_source(
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
        column_name: str,
    ) -> set[str]:
        """Creates a set of unique parent workspace IDs."""
        set_of_ids: set[str] = {
            getattr(workspace, column_name)
            for workspace in source_group
            if getattr(workspace, column_name)
        }
        return set_of_ids

    @staticmethod
    def get_set_of_upstream_workspace_ids(
        upstream_group: list[CatalogWorkspace],
    ) -> set[str]:
        """Creates a set of unique upstream workspace IDs."""
        set_of_ids: set[str] = {item.workspace_id for item in upstream_group}
        return set_of_ids

    def _get_child_to_wdfs_map(
        self,
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
    ) -> dict[str, dict[str, list[str]]]:
        """Creates a map of child workspace IDs to their WDF IDs."""
        # TODO: Use objects or a more transparent data structure instead of this.
        child_to_wdf_map: dict[str, dict[str, list[str]]] = {}

        # For each child, get its possible WDF IDs and values for each id
        for workspace in source_group:
            child_id: str = workspace.workspace_id
            wdf_id: str | None = workspace.workspace_data_filter_id
            wdf_values: list[str] | None = (
                workspace.workspace_data_filter_values
            )

            if wdf_values and wdf_id:
                if not child_to_wdf_map.get(child_id):
                    child_to_wdf_map[child_id] = {}
                child_to_wdf_map[child_id][wdf_id] = wdf_values

        return child_to_wdf_map

    def set_maps_based_on_source(
        self,
        map_object: WorkspaceDataMaps,
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
    ) -> WorkspaceDataMaps:
        """Creates maps which are dependent on the source group only."""
        map_object.child_to_parent_id_map = self._get_child_to_parent_map(
            source_group
        )
        map_object.workspace_id_to_wdf_map = self._get_child_to_wdfs_map(
            source_group
        )
        map_object.parent_ids = self._get_set_of_ids_from_source(
            source_group, "parent_id"
        )
        map_object.source_ids = self._get_set_of_ids_from_source(
            source_group, "workspace_id"
        )

        return map_object

    def set_maps_with_upstream_data(
        self,
        map_object: WorkspaceDataMaps,
        source_group: list[WorkspaceFullLoad] | list[WorkspaceIncrementalLoad],
        upstream_group: list[CatalogWorkspace],
    ) -> WorkspaceDataMaps:
        """Creates maps which are dependent on both the source group and upstream group."""
        map_object.workspace_id_to_name_map = self._get_id_to_name_map(
            source_group, upstream_group
        )
        map_object.upstream_ids = self.get_set_of_upstream_workspace_ids(
            upstream_group
        )

        return map_object



# (C) 2025 GoodData Corporation

from pathlib import Path
from unittest.mock import Mock

import boto3
import pytest
from moto import mock_aws
from pytest_mock import MockerFixture

from gooddata_pipelines import (
    PermissionProvisioner,
    UserDataFilterProvisioner,
    UserGroupProvisioner,
    UserProvisioner,
    WorkspaceProvisioner,
)
from gooddata_pipelines.api import GoodDataApi

TEST_DATA_DIR = str((Path(__file__).parent / "data").absolute())


@pytest.fixture(scope="function", autouse=True)
def mock_aws_services():
    """Mock AWS services for each test"""
    with mock_aws():
        yield


@pytest.fixture(scope="function")
def mock_boto_session():
    """Create a mock boto3 session for testing"""
    return boto3.Session(
        aws_access_key_id="testing",
        aws_secret_access_key="testing",
        aws_session_token="testing",
        region_name="us-east-1",
    )


@pytest.fixture(scope="function", autouse=True)
def mock_boto3_session(monkeypatch):
    """Mock boto3.Session to prevent real AWS authentication"""
    mock_client = Mock()
    mock_client.head_bucket.return_value = {}

    mock_resource = Mock()
    mock_resource.meta.client = mock_client
    mock_resource.Bucket.return_value = Mock()

    mock_session = Mock()
    mock_session.resource.return_value = mock_resource

    # Mock the boto3.Session constructor
    def mock_session_constructor(*args, **kwargs):
        return mock_session

    monkeypatch.setattr("boto3.Session", mock_session_constructor)
    return mock_session


@pytest.fixture
def mock_gooddata_api():
    return GoodDataApi("domain", "token")


@pytest.fixture
def mock_logger():
    class MockLogger:
        def info(self, msg: str) -> None:
            print(msg)

        def warning(self, msg: str) -> None:
            print(msg)

        def error(self, msg: str) -> None:
            print(msg)

    return MockLogger()


@pytest.fixture
def workspace_provisioner(mocker: MockerFixture):
    provisioner_instance = WorkspaceProvisioner.create("host", "token")

    # Patch the API
    mocker.patch.object(provisioner_instance, "_api", return_value=None)

    return provisioner_instance


@pytest.fixture
def user_provisioner(mocker: MockerFixture):
    provisioner_instance = UserProvisioner.create("host", "token")

    # Patch the API
    mocker.patch.object(provisioner_instance, "_api", return_value=None)

    return UserProvisioner.create("host", "token")


@pytest.fixture
def user_group_provisioner(mocker: MockerFixture):
    provisioner_instance = UserGroupProvisioner.create("host", "token")

    # Patch the API
    mocker.patch.object(provisioner_instance, "_api", return_value=None)

    return provisioner_instance


@pytest.fixture
def permission_provisioner(mocker: MockerFixture) -> PermissionProvisioner:
    provisioner_instance = PermissionProvisioner.create("host", "token")

    # Patch the API
    mocker.patch.object(provisioner_instance, "_api", return_value=None)

    return provisioner_instance


@pytest.fixture
def user_data_filter_provisioner(mocker: MockerFixture):
    provisioner_instance = UserDataFilterProvisioner.create("host", "token")

    # Patch the API
    mocker.patch.object(provisioner_instance, "_api", return_value=None)

    return provisioner_instance



# (C) 2025 GoodData Corporation

"""Interaction with GoodData Cloud via the direct API calls."""

import json
from typing import Any

import requests

TIMEOUT = 60
REQUEST_PAGE_SIZE = 250
API_VERSION = "v1"


class ApiMethods:
    headers: dict[str, str]
    base_url: str

    @staticmethod
    def _get_base_url(domain: str) -> str:
        """Returns the root endpoint for the GoodData Cloud API.

        Method ensures that the URL starts with "https://" and does not
        end with a trailing slash.

        Args:
            domain (str): The domain of the GoodData Cloud instance.
        Returns:
            str: The base URL for the GoodData Cloud API.
        """
        # Remove trailing slash if present.
        if domain[-1] == "/":
            domain = domain[:-1]

        if not domain.startswith("https://") and not domain.startswith(
            "http://"
        ):
            domain = f"https://{domain}"

        if domain.startswith("http://") and not domain.startswith("https://"):
            domain = domain.replace("http://", "https://")

        return f"{domain}/api/{API_VERSION}"

    def _get_url(self, endpoint: str) -> str:
        """Returns the full URL for a given API endpoint.

        Args:
            endpoint (str): The API endpoint to be appended to the base URL.
        Returns:
            str: The full URL for the API endpoint.
        """
        return f"{self.base_url}{endpoint}"

    def get_all_workspace_data_filters(
        self, workspace_id: str
    ) -> requests.Response:
        """Gets all workspace data filters for a given workspace.

        Args:
            workspace_id (str): The ID of the workspace.
        Returns:
            requests.Response: The response from the server containing all
                workspace data filters.
        """
        url = f"/entities/workspaces/{workspace_id}/workspaceDataFilters"
        return self._get(url)

    def get_workspace_data_filter_settings(
        self, workspace_id: str
    ) -> requests.Response:
        """Gets all workspace data filter settings for a given workspace.

        Args:
            workspace_id (str): The ID of the workspace.
        Returns:
            requests.Response: The response from the server containing all
                workspace data filter settings.
        """
        url = f"/entities/workspaces/{workspace_id}/workspaceDataFilterSettings?include=workspaceDataFilters"
        return self._get(url)

    def get_workspace_data_filter_setting(
        self, workspace_id: str, wdf_id: str
    ) -> requests.Response:
        """Gets a specific workspace data filter setting.

        Args:
            workspace_id (str): The ID of the workspace.
            wdf_id (str): The ID of the workspace data filter setting.
        Returns:
            requests.Response: The response from the server containing the
                workspace data filter setting.
        """
        url = f"/entities/workspaces/{workspace_id}/workspaceDataFilterSettings/{wdf_id}"
        return self._get(url)

    def put_workspace_data_filter_setting(
        self,
        workspace_id: str,
        wdf_setting: dict[str, Any],
    ) -> requests.Response:
        """Updates a workspace data filter setting.

        Args:
            workspace_id (str): The ID of the workspace.
            wdf_setting (dict[str, Any]): The workspace data filter setting to
                update.
        Returns:
            requests.Response: The response from the server containing the
                updated workspace data filter setting.
        """
        wdf_setting_id = wdf_setting["data"]["id"]
        endpoint = f"/entities/workspaces/{workspace_id}/workspaceDataFilterSettings/{wdf_setting_id}"
        return self._put(
            endpoint,
            wdf_setting,
            self.headers,
        )

    def post_workspace_data_filter_setting(
        self,
        workspace_id: str,
        wdf_setting: dict[str, Any],
    ) -> requests.Response:
        """Creates a workspace data filter setting for a given workspace.

        Args:
            workspace_id (str): The ID of the workspace.
            wdf_setting (dict[str, Any]): The workspace data filter setting to
                create.
        Returns:
            requests.Response: The response from the server containing the
                created workspace data filter setting.
        """
        endpoint = (
            f"/entities/workspaces/{workspace_id}/workspaceDataFilterSettings/"
        )
        return self._post(
            endpoint,
            wdf_setting,
            self.headers,
        )

    def delete_workspace_data_filter_setting(
        self,
        workspace_id: str,
        wdf_setting_id: str,
    ) -> requests.Response:
        """Deletes a workspace data filter setting.

        Args:
            workspace_id (str): The ID of the workspace.
            wdf_setting_id (str): The ID of the workspace data filter setting
                to delete.
        Returns:
            requests.Response: The response from the server confirming the
                deletion of the workspace data filter setting.
        """
        endpoint = f"/entities/workspaces/{workspace_id}/workspaceDataFilterSettings/{wdf_setting_id}"
        return self._delete(
            endpoint,
        )

    def get_user_data_filters(self, workspace_id: str) -> requests.Response:
        """Gets the user data filters for a given workspace."""
        endpoint = f"/layout/workspaces/{workspace_id}/userDataFilters"
        return self._get(endpoint)

    def put_user_data_filters(
        self, workspace_id: str, user_data_filters: dict[str, Any]
    ) -> requests.Response:
        """Puts the user data filters into GoodData workspace."""
        headers = {**self.headers, "Content-Type": "application/json"}
        return self._put(
            f"/layout/workspaces/{workspace_id}/userDataFilters",
            user_data_filters,
            headers,
        )

    def get_automations(self, workspace_id: str) -> requests.Response:
        """Gets the automations for a given workspace."""
        endpoint = (
            f"/entities/workspaces/{workspace_id}/automations?include=ALL"
        )
        return self._get(endpoint)

    def post_automation(
        self, workspace_id: str, automation: dict[str, Any]
    ) -> requests.Response:
        """Posts an automation for a given workspace."""
        endpoint = f"/entities/workspaces/{workspace_id}/automations"
        return self._post(endpoint, automation)

    def delete_automation(
        self, workspace_id: str, automation_id: str
    ) -> requests.Response:
        """Deletes an automation for a given workspace."""
        endpoint = (
            f"/entities/workspaces/{workspace_id}/automations/{automation_id}"
        )
        return self._delete(endpoint)

    def get_all_metrics(self, workspace_id: str) -> requests.Response:
        """Get all metrics from the specified workspace.

        Args:
            workspace_id (str): The ID of the workspace to retrieve metrics from.
        Returns:
            requests.Response: The response containing the metrics.
        """
        endpoint = f"/entities/workspaces/{workspace_id}/metrics"
        headers = {**self.headers, "X-GDC-VALIDATE-RELATIONS": "true"}
        return self._get(endpoint, headers=headers)

    def get_all_visualization_objects(
        self, workspace_id: str
    ) -> requests.Response:
        """Get all visualizations from the specified workspace.

        Args:
            workspace_id (str): The ID of the workspace to retrieve visualizations from.
        Returns:
            requests.Response: The response containing the visualizations.
        """
        endpoint = f"/entities/workspaces/{workspace_id}/visualizationObjects"
        headers = {**self.headers, "X-GDC-VALIDATE-RELATIONS": "true"}
        return self._get(endpoint, headers=headers)

    def get_all_dashboards(self, workspace_id: str) -> requests.Response:
        """Get all dashboards from the specified workspace.

        Args:
            workspace_id (str): The ID of the workspace to retrieve dashboards from.
        Returns:
            requests.Response: The response containing the dashboards.
        """
        endpoint = f"/entities/workspaces/{workspace_id}/analyticalDashboards"
        headers = {**self.headers, "X-GDC-VALIDATE-RELATIONS": "true"}
        return self._get(endpoint, headers=headers)

    def get_profile(self) -> requests.Response:
        """Returns organization and current user information."""
        endpoint = "/profile"
        return self._get(endpoint)

    def _get(
        self, endpoint: str, headers: dict[str, str] | None = None
    ) -> requests.Response:
        """Sends a GET request to the server.

        Args:
            endpoint (str): The API endpoint to send the GET request to.
            headers (dict[str, str] | None): Headers to include in the request.
                If no headers are provided, the default headers will be used.
        Returns:
            requests.Response: The response from the server.
        """
        url = self._get_url(endpoint)
        request_headers = headers if headers else self.headers

        return requests.get(url, headers=request_headers, timeout=TIMEOUT)

    def _post(
        self,
        endpoint: str,
        data: Any,
        headers: dict | None = None,
    ) -> requests.Response:
        """Sends a POST request to the server with a given JSON object.

        Args:
            endpoint (str): The API endpoint to send the POST request to.
            data (Any): The JSON data to send in the request body.
            headers (dict | None): Headers to include in the request.
                If no headers are provided, the default headers will be used.
        Returns:
            requests.Response: The response from the server.
        """
        url = self._get_url(endpoint)
        request_headers = headers if headers else self.headers
        data_json = json.dumps(data)

        return requests.post(
            url, data=data_json, headers=request_headers, timeout=TIMEOUT
        )

    def _put(
        self,
        endpoint: str,
        data: Any,
        headers: dict | None = None,
    ) -> requests.Response:
        """Sends a PUT request to the server with a given JSON object.

        Args:
            endpoint (str): The API endpoint to send the PUT request to.
            data (Any): The JSON data to send in the request body.
            headers (dict | None): Headers to include in the request.
                If no headers are provided, the default headers will be used.
        Returns:
            requests.Response: The response from the server.
        """
        url = self._get_url(endpoint)
        request_headers = headers if headers else self.headers
        data_json = json.dumps(data)

        return requests.put(
            url, data=data_json, headers=request_headers, timeout=TIMEOUT
        )

    def _delete(
        self,
        endpoint: str,
    ) -> requests.Response:
        """Sends a DELETE request to the server.

        Args:
            endpoint (str): The API endpoint to send the DELETE request to.
        Returns:
            requests.Response: The response from the server.
        """
        url = self._get_url(endpoint)

        return requests.delete(url, headers=self.headers, timeout=TIMEOUT)

    @staticmethod
    def raise_if_response_not_ok(*responses: requests.Response) -> None:
        """Check if responses from API calls are OK.

        Raises ValueError if any response is not OK (status code not 2xx).
        """
        for response in responses:
            if not response.ok:
                raise ValueError(
                    f"Request to {response.url} failed with status code {response.status_code}: {response.text}"
                )



# (C) 2025 GoodData Corporation

from typing import Any

from gooddata_pipelines.logger import LoggerLike
from gooddata_pipelines.provisioning.generic.config import (
    PROVISIONING_CONFIG,
    LoadType,
    WorkflowType,
)


def provision(
    data: list[dict[str, Any]],
    workflow_type: WorkflowType,
    host: str,
    token: str,
    logger: LoggerLike | None = None,
) -> None:
    """Generic provisioning function accepting raw data and workflow type.

    The function will validate data based on the selected workflow type and run
    the corresponding provisioning in full or incremental mode.

    Args:
        data: List of dictionaries containing the data to be provisioned.
        workflow_type: The type of workflow to run.
        host: The host of the GoodData platform.
        token: The token for the GoodData platform.
        logger: The logger to use for logging.
    """

    if workflow_type not in PROVISIONING_CONFIG:
        raise ValueError(f"Invalid workflow type: {workflow_type}")

    config = PROVISIONING_CONFIG[workflow_type]

    provisioner = config.provisioner_class.create(host, token)
    validated_data: list = [config.validation_model(**item) for item in data]

    if logger:
        provisioner.logger.subscribe(logger)

    if config.load_type == LoadType.FULL:
        provisioner.full_load(validated_data)
    elif config.load_type == LoadType.INCREMENTAL:
        provisioner.incremental_load(validated_data)
    else:
        raise ValueError(f"Invalid load type: {config.load_type}")



# (C) 2025 GoodData Corporation
"""Module for processing validated custom datasets and fields data.

This module is responsible for converting validated custom datasets and fields
into objects defined in the GoodData Python SDK.
"""

from gooddata_sdk.catalog.identifier import (
    CatalogDatasetWorkspaceDataFilterIdentifier,
    CatalogGrainIdentifier,
    CatalogReferenceIdentifier,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.logical_model.data_filter_references import (
    CatalogDeclarativeWorkspaceDataFilterReferences,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.logical_model.dataset.dataset import (
    CatalogDataSourceTableIdentifier,
    CatalogDeclarativeAttribute,
    CatalogDeclarativeDataset,
    CatalogDeclarativeDatasetSql,
    CatalogDeclarativeFact,
    CatalogDeclarativeReference,
    CatalogDeclarativeReferenceSource,
    CatalogDeclarativeWorkspaceDataFilterColumn,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.logical_model.date_dataset.date_dataset import (
    CatalogDeclarativeDateDataset,
    CatalogGranularitiesFormatting,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.logical_model.ldm import (
    CatalogDeclarativeLdm,
    CatalogDeclarativeModel,
)

from gooddata_pipelines.ldm_extension.models.aliases import DatasetId
from gooddata_pipelines.ldm_extension.models.custom_data_object import (
    ColumnDataType,
    CustomDataset,
    CustomFieldDefinition,
    CustomFieldType,
)


class LdmExtensionDataProcessor:
    """Create GoodData LDM from validated custom datasets and fields."""

    DATE_GRANULARITIES: list[str] = [
        "MINUTE",
        "HOUR",
        "DAY",
        "WEEK",
        "MONTH",
        "QUARTER",
        "YEAR",
        "MINUTE_OF_HOUR",
        "HOUR_OF_DAY",
        "DAY_OF_WEEK",
        "DAY_OF_MONTH",
        "DAY_OF_YEAR",
        "WEEK_OF_YEAR",
        "MONTH_OF_YEAR",
        "QUARTER_OF_YEAR",
    ]

    @staticmethod
    def _attribute_from_field(
        dataset_name: str,
        custom_field: CustomFieldDefinition,
    ) -> CatalogDeclarativeAttribute:
        """Assign a declarative attribute from a custom field definition."""
        return CatalogDeclarativeAttribute(
            id=custom_field.custom_field_id,
            title=custom_field.custom_field_name,
            source_column=custom_field.custom_field_source_column,
            labels=[],
            source_column_data_type=custom_field.custom_field_source_column_data_type.value,
            tags=[dataset_name],
        )

    @staticmethod
    def _fact_from_field(
        dataset_name: str,
        custom_field: CustomFieldDefinition,
    ) -> CatalogDeclarativeFact:
        """Assign a declarative fact from a custom field definition."""
        return CatalogDeclarativeFact(
            id=custom_field.custom_field_id,
            title=custom_field.custom_field_name,
            source_column=custom_field.custom_field_source_column,
            source_column_data_type=custom_field.custom_field_source_column_data_type.value,
            tags=[dataset_name],
        )

    def _date_from_field(
        self,
        dataset_name: str,
        custom_field: CustomFieldDefinition,
    ) -> CatalogDeclarativeDateDataset:
        """Assign a declarative date dataset from a custom field definition."""

        return CatalogDeclarativeDateDataset(
            id=custom_field.custom_field_id,
            title=custom_field.custom_field_name,
            granularities_formatting=CatalogGranularitiesFormatting(
                title_base="",
                title_pattern="%titleBase - %granularityTitle",
            ),
            granularities=self.DATE_GRANULARITIES,
            tags=[dataset_name],
        )

    @staticmethod
    def _date_ref_from_field(
        custom_field: CustomFieldDefinition,
    ) -> CatalogDeclarativeReference:
        """Create a date reference from a custom field definition."""
        return CatalogDeclarativeReference(
            identifier=CatalogReferenceIdentifier(
                id=custom_field.custom_field_id
            ),
            multivalue=False,
            sources=[
                CatalogDeclarativeReferenceSource(
                    column=custom_field.custom_field_source_column,
                    target=CatalogGrainIdentifier(
                        id=custom_field.custom_field_id,
                        type=CustomFieldType.DATE.value,
                    ),
                    data_type=custom_field.custom_field_source_column_data_type.value,
                )
            ],
        )

    @staticmethod
    def _get_sources(
        dataset: CustomDataset,
    ) -> tuple[
        CatalogDataSourceTableIdentifier | None,
        CatalogDeclarativeDatasetSql | None,
    ]:
        """Get the data source table and SQL from the dataset definition."""
        # We will have either a table id or a sql statement. Let's store
        # whatever data is available to variables and pass it to the
        # dataset. Both can be object instances or None, but at least one
        # should be valid as per prior validation.
        dataset_source_table_id = (
            CatalogDataSourceTableIdentifier(
                id=dataset.definition.dataset_source_table,
                data_source_id=dataset.definition.dataset_datasource_id,
                path=[dataset.definition.dataset_source_table],
            )
            if dataset.definition.dataset_source_table
            else None
        )

        dataset_sql = (
            CatalogDeclarativeDatasetSql(
                statement=dataset.definition.dataset_source_sql,
                data_source_id=dataset.definition.dataset_datasource_id,
            )
            if dataset.definition.dataset_source_sql
            else None
        )
        return dataset_source_table_id, dataset_sql

    def datasets_to_ldm(
        self, datasets: dict[DatasetId, CustomDataset]
    ) -> CatalogDeclarativeModel:
        """Convert validated datasets to GoodData declarative model.

        Args:
            datasets (dict[DatasetId, CustomDataset]): Dictionary of validated
                datasets.
        Returns:
            CatalogDeclarativeModel: GoodData declarative model representation
                of the datasets.
        """

        declarative_datasets: list[CatalogDeclarativeDataset] = []

        # Date dimensions are not stored in a dataset, but as a separate datasets
        # in `date_instances` object on the LDM
        date_instances: list[CatalogDeclarativeDateDataset] = []

        for dataset in datasets.values():
            date_references: list[CatalogDeclarativeReference] = []
            attributes: list[CatalogDeclarativeAttribute] = []
            facts: list[CatalogDeclarativeFact] = []

            # Iterate through the custom fields and create the appropriate objects
            for custom_field in dataset.custom_fields:
                if custom_field.custom_field_type == CustomFieldType.ATTRIBUTE:
                    attributes.append(
                        self._attribute_from_field(
                            dataset.definition.dataset_name, custom_field
                        )
                    )

                elif custom_field.custom_field_type == CustomFieldType.FACT:
                    facts.append(
                        self._fact_from_field(
                            dataset.definition.dataset_name, custom_field
                        )
                    )

                # Process date dimensions and store them to date_instances. Date
                # dimensions are not stored in a dataset, but as a separate dataset.
                # However, they need to be referenced in the dataset references to
                # create the connection between the dataset and the date dimension
                # in the GoodData Logical Data Model.
                elif custom_field.custom_field_type == CustomFieldType.DATE:
                    # Add the date dimension to the date_instances
                    date_instances.append(
                        self._date_from_field(
                            dataset.definition.dataset_name, custom_field
                        )
                    )

                    # Create a reference so that the date dimension is connected
                    # to the dataset in the GoodData Logical Data Model.
                    date_references.append(
                        self._date_ref_from_field(custom_field)
                    )

                else:
                    raise ValueError(
                        f"Unsupported custom field type: {custom_field.custom_field_type}"
                    )

            # Get the data source info
            dataset_source_table_id, dataset_sql = self._get_sources(dataset)

            # Construct the declarative dataset object and append it to the list.
            declarative_datasets.append(
                CatalogDeclarativeDataset(
                    id=dataset.definition.dataset_id,
                    title=dataset.definition.dataset_name,
                    grain=[],
                    references=[
                        CatalogDeclarativeReference(
                            identifier=CatalogReferenceIdentifier(
                                id=dataset.definition.parent_dataset_reference,
                            ),
                            multivalue=True,
                            sources=[
                                CatalogDeclarativeReferenceSource(
                                    column=dataset.definition.dataset_reference_source_column,
                                    data_type=dataset.definition.dataset_reference_source_column_data_type.value,
                                    target=CatalogGrainIdentifier(
                                        id=dataset.definition.parent_dataset_reference_attribute_id,
                                        type=CustomFieldType.ATTRIBUTE.value,
                                    ),
                                )
                            ],
                        ),
                    ]
                    + date_references,
                    description=None,
                    attributes=attributes,
                    facts=facts,
                    data_source_table_id=dataset_source_table_id,
                    sql=dataset_sql,
                    workspace_data_filter_columns=[
                        CatalogDeclarativeWorkspaceDataFilterColumn(
                            name=dataset.definition.workspace_data_filter_column_name,
                            data_type=ColumnDataType.STRING.value,
                        )
                    ],
                    workspace_data_filter_references=[
                        CatalogDeclarativeWorkspaceDataFilterReferences(
                            filter_id=CatalogDatasetWorkspaceDataFilterIdentifier(
                                id=dataset.definition.workspace_data_filter_id
                            ),
                            filter_column=dataset.definition.workspace_data_filter_column_name,
                            filter_column_data_type=ColumnDataType.STRING.value,
                        )
                    ],
                    tags=[dataset.definition.dataset_name],
                )
            )

        # Create the Logical Data Model from the datasets and the date instances.
        ldm = CatalogDeclarativeLdm(
            datasets=declarative_datasets, date_instances=date_instances
        )
        return CatalogDeclarativeModel(ldm=ldm)



# (C) 2025 GoodData Corporation

from enum import Enum


class InputType(Enum):
    """Input type for the backup."""

    LIST_OF_WORKSPACES = "list-of-workspaces"
    HIERARCHY = "list-of-parents"
    ORGANIZATION = "entire-organization"



# (C) 2025 GoodData Corporation

"""Module for context objects used in GoodData Pipelines provisioning."""


class WorkspaceContext:
    workspace_id: str
    workspace_name: str | None
    wdf_id: str | None
    wdf_values: list[str] | None

    def __init__(
        self,
        workspace_id: str | None,
        workspace_name: str | None,
        wdf_id: str | None = None,
        wdf_values: list[str] | None = None,
    ):
        self.workspace_id = workspace_id if workspace_id else "NA"
        self.workspace_name = workspace_name
        self.wdf_id = wdf_id
        self.wdf_values = wdf_values


class UserContext:
    user_id: str
    user_groups: str

    def __init__(self, user_id: str, user_groups: list[str]):
        """User context object, stringifies list of user groups"""
        self.user_id = user_id
        self.user_groups = ",".join(user_groups)



# (C) 2025 GoodData Corporation

import time
import pytest
from gooddata_pipelines.utils.rate_limiter import RateLimiter


# ---------------------------
# Core wait + reset behavior
# ---------------------------


def test_rate_limiter_no_wait_needed():
    limiter = RateLimiter(calls_per_second=1000.0)  # Very fast limit
    waited = limiter.wait_if_needed()
    assert waited == pytest.approx(0.0, abs=0.001)


def test_rate_limiter_enforces_delay():
    limiter = RateLimiter(calls_per_second=2.0)
    limiter.wait_if_needed()
    start = time.time()
    waited = limiter.wait_if_needed()
    duration = time.time() - start

    assert waited >= 0.49
    assert duration < 0.65


def test_rate_limiter_respects_reset():
    limiter = RateLimiter(calls_per_second=1.0)
    limiter.wait_if_needed()
    limiter.reset()
    waited = limiter.wait_if_needed()
    assert waited == pytest.approx(0.0, abs=0.001)


def test_rate_limiter_min_interval_property():
    limiter = RateLimiter(calls_per_second=4.0)
    assert limiter.min_interval == pytest.approx(0.25, abs=1e-9)


# -----------------------------------------
# Decorator: shared instance (@limiter)
# -----------------------------------------


def test_rate_limiter_as_decorator_enforces_delay_shared_instance():
    limiter = RateLimiter(calls_per_second=2.0)
    ts = []

    @limiter
    def func():
        ts.append(time.time())

    func()
    func()

    assert len(ts) == 2
    assert ts[1] - ts[0] >= 0.49


def test_rate_limiter_decorator_shared_state_across_functions():
    limiter = RateLimiter(calls_per_second=2.0)
    ts = []

    @limiter
    def func_a():
        ts.append(time.time())

    @limiter
    def func_b():
        ts.append(time.time())

    func_a()
    func_b()  # should be throttled by the *same* limiter
    assert len(ts) == 2
    assert ts[1] - ts[0] >= 0.49


def test_multiple_limiters_independent_state_shared_instance_mode():
    limiter_a = RateLimiter(calls_per_second=2.0)
    limiter_b = RateLimiter(calls_per_second=2.0)

    ts_a = []
    ts_b = []

    @limiter_a
    def func_a():
        ts_a.append(time.time())

    @limiter_b
    def func_b():
        ts_b.append(time.time())

    func_a()
    func_b()

    # They should be ~simultaneous since they use different instances
    assert abs(ts_a[0] - ts_b[0]) < 0.05


# -------------------------------------------------------
# Decorator: per-function instance (@RateLimiter(...))
# -------------------------------------------------------


def test_per_function_decorator_enforces_delay_per_function():
    # Each function decorated this way gets its *own* limiter instance.
    ts = []

    @RateLimiter(calls_per_second=2.0)  # 0.5s
    def func():
        ts.append(time.time())

    func()
    func()
    assert len(ts) == 2
    assert ts[1] - ts[0] >= 0.49


def test_per_function_decorator_independent_state_between_functions():
    ts_a = []
    ts_b = []

    @RateLimiter(calls_per_second=2.0)
    def func_a():
        ts_a.append(time.time())

    @RateLimiter(calls_per_second=2.0)
    def func_b():
        ts_b.append(time.time())

    func_a()
    func_b()  # independent limiter, so no enforced delay between A and B
    assert abs(ts_a[0] - ts_b[0]) < 0.05


# -----------------------------
# Context manager usage
# -----------------------------


def test_context_manager_waits_on_enter():
    limiter = RateLimiter(calls_per_second=2.0)
    with limiter:
        t1 = time.time()
    with limiter:
        t2 = time.time()

    # The second 'with' should be at least 0.5s after the first 'with' enter
    assert t2 - t1 >= 0.49


def test_context_manager_multiple_uses_same_instance():
    limiter = RateLimiter(calls_per_second=3.0)
    times = []

    for _ in range(3):
        with limiter:
            times.append(time.time())

    intervals = [b - a for a, b in zip(times, times[1:])]
    for iv in intervals:
        assert iv >= 0.30  # a bit of slack for timing variance


def test_context_manager_propagates_exceptions():
    limiter = RateLimiter(calls_per_second=10.0)

    class Boom(Exception):
        pass

    with pytest.raises(Boom):
        with limiter:
            raise Boom("fail")



# (C) 2025 GoodData Corporation

import json
import os
import shutil
import tempfile
from pathlib import Path

import pytest

from gooddata_pipelines.backup_and_restore.constants import DirNames
from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
    S3StorageConfig,
    StorageType,
)
from gooddata_pipelines.backup_and_restore.restore_manager import (
    RestoreManager,
    WorkspaceModel,
    WorkspaceToRestore,
)
from gooddata_pipelines.backup_and_restore.storage.s3_storage import S3Storage
from tests.conftest import TEST_DATA_DIR

TEST_DATA_SUBDIR = f"{TEST_DATA_DIR}/restore"

MOCK_DL_TARGET = Path("overlays.zip")
TEST_CONF_PATH = f"{TEST_DATA_SUBDIR}/test_conf.yaml"
TEST_UDF_ROOT = Path(f"{TEST_DATA_SUBDIR}/test_udf_root").absolute()

S3_BACKUP_PATH = "some/s3/backup/path/org_id/"
S3_BUCKET = "some-s3-bucket"


# Small reusable fixtures to avoid repeated mocking
@pytest.fixture()
def conf_s3() -> BackupRestoreConfig:
    return BackupRestoreConfig.from_yaml(TEST_CONF_PATH)


@pytest.fixture()
def s3_storage(mocker, conf_s3: BackupRestoreConfig) -> S3Storage:
    mocker.patch.object(S3Storage, "_verify_connection", return_value=None)
    storage = S3Storage(conf_s3)
    return storage


@pytest.fixture()
def s3_bucket(mocker, s3_storage: S3Storage):
    bucket = mocker.MagicMock()
    s3_storage._bucket = bucket  # type: ignore[attr-defined]
    return bucket


@pytest.fixture()
def gd_api_instance(mocker):
    gd_api_cls = mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.GoodDataApi",
    )
    gd_api_instance = gd_api_cls.return_value
    gd_api_instance._sdk = mocker.MagicMock()
    return gd_api_instance


@pytest.fixture()
def restore_manager(mocker, gd_api_instance) -> RestoreManager:  # noqa: ARG001
    mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.S3Storage",
        autospec=True,
    )
    config = BackupRestoreConfig(
        storage_type=StorageType.S3,
        storage=S3StorageConfig(backup_path=S3_BACKUP_PATH, bucket=S3_BUCKET),
    )
    manager = RestoreManager.create(host="host", token="token", config=config)
    return manager


def assert_not_called_with(target, *args, **kwargs):
    try:
        target.assert_called_with(*args, **kwargs)
    except AssertionError:
        return
    formatted_call = target._format_mock_call_signature(args, kwargs)
    raise AssertionError(f"Expected {formatted_call} to not have been called.")


def test_s3_storage_success(s3_storage, s3_bucket, mocker):
    """S3Storage: downloads the first zip object when objects are present."""
    dir_marker = mocker.MagicMock()
    zip_obj = mocker.MagicMock()
    zip_obj.key = f"{S3_BACKUP_PATH}ws_id/gooddata_layouts.zip"
    s3_bucket.objects.filter.return_value = [dir_marker, zip_obj]

    with tempfile.TemporaryDirectory() as tempdir:
        target_path = Path(tempdir, MOCK_DL_TARGET)
        s3_storage.get_ws_declaration("ws_id/", target_path)
        s3_bucket.download_file.assert_called_once_with(
            zip_obj.key, target_path
        )


def test_s3_storage_no_target_only_dir(s3_storage, s3_bucket, mocker):
    """S3Storage: raises when only a directory marker exists under the prefix."""
    s3_bucket.objects.filter.return_value = [mocker.MagicMock()]
    with pytest.raises(Exception):
        s3_storage.get_ws_declaration("ws_id/", MOCK_DL_TARGET)


def test_s3_storage_no_target(s3_storage, s3_bucket, mocker):
    """S3Storage: raises when no objects exist for the given prefix."""
    s3_bucket.objects.filter.return_value = []
    with pytest.raises(Exception):
        s3_storage.get_ws_declaration("bad_target/", MOCK_DL_TARGET)


def test_restore_empty_workspace(restore_manager, gd_api_instance, mocker):
    """RestoreManager: valid layout triggers LDM and AM PUTs."""
    mocker.patch.object(
        restore_manager.storage, "get_ws_declaration", return_value=None
    )

    def create_empty_ws(_, destination: Path):
        os.mkdir(destination / DirNames.LAYOUTS)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.LDM)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.AM)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.UDF)
        os.mkdir(destination / DirNames.LAYOUTS / "filter_views")
        os.mkdir(destination / DirNames.LAYOUTS / "automations")

    mocker.patch.object(
        restore_manager, "_extract_zip_archive", side_effect=create_empty_ws
    )

    workspace_model = WorkspaceModel(
        logical_data_model=mocker.Mock(), analytics_model=mocker.Mock()
    )
    mocker.patch.object(
        restore_manager, "_load_workspace_layout", return_value=workspace_model
    )
    mocker.patch.object(
        restore_manager,
        "_load_user_data_filters",
        return_value={"userDataFilters": []},
    )
    mocker.patch.object(
        restore_manager, "_load_and_put_filter_views", return_value=None
    )
    mocker.patch.object(
        restore_manager, "_load_and_post_automations", return_value=None
    )

    restore_manager.restore(
        [WorkspaceToRestore(id="ws_id", path="some/ws/path")]
    )

    gd_api_instance._sdk.catalog_workspace_content.put_declarative_ldm.assert_called_once_with(
        "ws_id", workspace_model.logical_data_model
    )
    gd_api_instance._sdk.catalog_workspace_content.put_declarative_analytics_model.assert_called_once_with(
        "ws_id", workspace_model.analytics_model
    )


def test_invalid_workspace_on_disk_is_skipped(
    restore_manager, gd_api_instance, mocker
):
    """RestoreManager: invalid layout (missing dirs) is skipped; no PUTs."""
    mocker.patch.object(
        restore_manager.storage, "get_ws_declaration", return_value=None
    )

    def create_invalid_ws(_, destination: Path):
        os.mkdir(destination / DirNames.LAYOUTS)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.LDM)
        # Missing AM and UDF

    mocker.patch.object(
        restore_manager, "_extract_zip_archive", side_effect=create_invalid_ws
    )

    restore_manager.restore(
        [WorkspaceToRestore(id="ws_id", path="some/ws/path")]
    )

    gd_api_instance._sdk.catalog_workspace_content.put_declarative_ldm.assert_not_called()
    gd_api_instance._sdk.catalog_workspace_content.put_declarative_analytics_model.assert_not_called()


def test_restore_multiple_workspaces_with_partial_failure(
    restore_manager, gd_api_instance, mocker
):
    """RestoreManager: multiple targets; on partial failure only successful PUTs occur."""
    ws_catalog = gd_api_instance._sdk.catalog_workspace_content

    mocker.patch.object(
        restore_manager.storage, "get_ws_declaration", return_value=None
    )

    def create_valid_ws(_, destination: Path):
        os.mkdir(destination / DirNames.LAYOUTS)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.LDM)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.AM)
        os.mkdir(destination / DirNames.LAYOUTS / DirNames.UDF)

    mocker.patch.object(
        restore_manager, "_extract_zip_archive", side_effect=create_valid_ws
    )

    workspace_model = WorkspaceModel(
        logical_data_model=mocker.Mock(), analytics_model=mocker.Mock()
    )
    # First load succeeds, second raises
    mocker.patch.object(
        restore_manager,
        "_load_workspace_layout",
        side_effect=[workspace_model, Exception()],
    )
    mocker.patch.object(
        restore_manager,
        "_load_user_data_filters",
        return_value={"userDataFilters": []},
    )

    targets = [
        WorkspaceToRestore(id="ws_id_1", path="ws_id_1"),
        WorkspaceToRestore(id="ws_id_2", path="ws_id_1"),
    ]
    restore_manager.restore(targets)

    ws_catalog.put_declarative_ldm.assert_any_call(
        "ws_id_1", workspace_model.logical_data_model
    )
    ws_catalog.put_declarative_analytics_model.assert_any_call(
        "ws_id_1", workspace_model.analytics_model
    )
    assert_not_called_with(
        ws_catalog.put_declarative_ldm, "ws_id_2", mocker.ANY
    )
    assert_not_called_with(
        ws_catalog.put_declarative_analytics_model, "ws_id_2", mocker.ANY
    )


def test_load_user_data_filters_reads_yaml(mocker):
    """RestoreManager: reads YAML UDFs into expected API body structure."""
    mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.S3Storage",
        autospec=True,
    )
    gd_api_cls = mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.GoodDataApi",
    )
    gd_api_instance = gd_api_cls.return_value
    gd_api_instance._sdk = mocker.MagicMock()

    config = BackupRestoreConfig(
        storage_type=StorageType.S3,
        storage=S3StorageConfig(backup_path=S3_BACKUP_PATH, bucket=S3_BUCKET),
    )
    manager = RestoreManager.create(host="host", token="token", config=config)

    # Build a clean temp directory and copy only YAML fixtures from test data
    with tempfile.TemporaryDirectory() as tempdir:
        temp_root = Path(tempdir)
        src_udf_dir = (TEST_UDF_ROOT / DirNames.UDF).absolute()
        dst_udf_dir = temp_root / DirNames.UDF

        def ignore_non_yaml(_dir, names):
            return [n for n in names if not n.endswith(".yaml")]

        shutil.copytree(src_udf_dir, dst_udf_dir, ignore=ignore_non_yaml)

        result = manager._load_user_data_filters(temp_root)

    user_data_filters_expected = {
        "userDataFilters": [
            {
                "id": "datafilter2",
                "maql": '{label/campaign_channels.category} = "1"',
                "title": "Status filter",
                "user": {
                    "id": "5c867a8a-12af-45bf-8d85-c7d16bedebd1",
                    "type": "user",
                },
            },
            {
                "id": "datafilter4",
                "maql": '{label/campaign_channels.category} = "1"',
                "title": "Status filter",
                "user": {
                    "id": "5c867a8a-12af-45bf-8d85-c7d16bedebd1",
                    "type": "user",
                },
            },
        ]
    }

    sorted_result = sorted(
        json.dumps(d, sort_keys=True) for d in result["userDataFilters"]
    )
    sorted_expected = sorted(
        json.dumps(d, sort_keys=True)
        for d in user_data_filters_expected["userDataFilters"]
    )

    assert sorted_result == sorted_expected


def test_manager_create_uses_s3_storage(mocker):
    """RestoreManager.create: builds S3 storage when storage_type is S3."""
    storage_cls = mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.S3Storage",
        autospec=True,
    )
    mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.GoodDataApi",
    )
    config = BackupRestoreConfig(
        storage_type=StorageType.S3,
        storage=S3StorageConfig(backup_path=S3_BACKUP_PATH, bucket=S3_BUCKET),
    )
    RestoreManager.create(host="host", token="token", config=config)
    storage_cls.assert_called_once()


def test_manager_create_from_profile(mocker):
    """RestoreManager.create_from_profile: uses profile content to init GoodDataApi."""
    mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.S3Storage",
        autospec=True,
    )
    gd_api_cls = mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.GoodDataApi",
    )
    mocker.patch(
        "gooddata_pipelines.backup_and_restore.base_manager.profile_content",
        return_value={"host": "h", "token": "t"},
    )

    config = BackupRestoreConfig(
        storage_type=StorageType.S3,
        storage=S3StorageConfig(backup_path=S3_BACKUP_PATH, bucket=S3_BUCKET),
    )
    RestoreManager.create_from_profile(config)
    gd_api_cls.assert_called_once_with("h", "t")



# (C) 2025 GoodData Corporation

import os

import boto3

from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
    S3StorageConfig,
)
from gooddata_pipelines.backup_and_restore.storage.base_storage import (
    BackupStorage,
)


class S3Storage(BackupStorage):
    def __init__(self, conf: BackupRestoreConfig):
        super().__init__(conf)

        if not isinstance(conf.storage, S3StorageConfig):
            raise ValueError("S3 storage config is required")

        self._config = conf.storage
        self._session = self._create_boto_session(self._config)
        self._client = self._session.client("s3")
        self._resource = self._session.resource("s3")
        self._bucket = self._resource.Bucket(self._config.bucket)  # type: ignore [missing library stubs]

        self._verify_connection()

    def _create_boto_session(self, config: S3StorageConfig) -> boto3.Session:
        if config.aws_access_key_id and config.aws_secret_access_key:
            if not config.aws_default_region:
                self.logger.warning(
                    "No AWS region specified. Defaulting to us-east-1."
                )
            try:
                return boto3.Session(
                    aws_access_key_id=config.aws_access_key_id,
                    aws_secret_access_key=config.aws_secret_access_key,
                    region_name=config.aws_default_region,
                )
            except Exception:
                self.logger.warning(
                    "Failed to create boto3 session with supplied credentials."
                )

        if config.profile:
            try:
                return boto3.Session(profile_name=config.profile)
            except Exception:
                self.logger.warning(
                    f"AWS profile [{config.profile}] not found."
                )

        try:
            return boto3.Session()
        except Exception:
            self.logger.error(
                "Failed to create boto3 session with default IAM role or environment credentials."
            )
            raise RuntimeError(
                "Unable to create AWS session. Please check your AWS credentials, profile, or IAM role configuration."
            )

    def _verify_connection(self) -> None:
        """
        Pings the S3 bucket to verify that the connection is working.
        """
        try:
            self._client.head_bucket(Bucket=self._config.bucket)
        except Exception as e:
            raise RuntimeError(
                f"Failed to connect to S3 bucket {self._config.bucket}: {e}"
            )

    def export(self, folder: str, org_id: str) -> None:
        """Uploads the content of the folder to S3 as a backup."""
        storage_path = f"{self._config.bucket}/{self._backup_path}"
        self.logger.info(f"Uploading {org_id} to {storage_path}")
        folder = f"{folder}/{org_id}"
        for subdir, dirs, files in os.walk(folder):
            full_path = os.path.join(subdir)
            export_path = (
                f"{self._backup_path}{org_id}/{full_path[len(folder) + 1 :]}/"
            )
            self._client.put_object(Bucket=self._config.bucket, Key=export_path)

            for file in files:
                full_path = os.path.join(subdir, file)
                with open(full_path, "rb") as data:
                    export_path = f"{self._backup_path}{org_id}/{full_path[len(folder) + 1 :]}"
                    self._client.put_object(
                        Bucket=self._config.bucket, Key=export_path, Body=data
                    )

    def get_ws_declaration(
        self, target_path: str, local_target_path: str
    ) -> None:
        """Retrieves workspace declaration from S3 bucket."""
        target_s3_prefix = f"{self._backup_path}{target_path}"

        objs_found = list(self._bucket.objects.filter(Prefix=target_s3_prefix))

        # Remove the included directory (which equals prefix) on hit
        objs_found = objs_found[1:] if len(objs_found) > 0 else objs_found

        if not objs_found:
            message = f"No target backup found for {target_s3_prefix}."
            self.logger.error(message)
            raise Exception(message)

        if len(objs_found) > 1:
            self.logger.warning(
                f"Multiple backups found at {target_s3_prefix}."
                " Continuing with the first one, ignoring the rest..."
            )

        s3_obj = objs_found[0]
        self._bucket.download_file(s3_obj.key, local_target_path)



# (C) 2025 GoodData Corporation

from typing import Any

import orjson
import pytest
from pydantic import ValidationError
from requests import Response

from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_filters import (
    WDFSetting,
    WorkspaceDataFilterManager,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_parser import (
    WorkspaceDataMaps,
)
from gooddata_pipelines.provisioning.utils.context_objects import (
    WorkspaceContext,
)
from gooddata_pipelines.provisioning.utils.exceptions import WorkspaceException
from tests.data.mock_responses import (
    WDF_ACTUAL_WDF_SETTINGS,
    WDF_VALID_PAYLOAD,
)


@pytest.fixture
def wdf_manager(mock_gooddata_api):
    return WorkspaceDataFilterManager(mock_gooddata_api, WorkspaceDataMaps())


def test_create_wdf_setting_dict(wdf_manager) -> None:
    """Test construction of the WDF setting dictionary."""
    wdf_setting_id: str = "expected_wdf_setting_id"
    wdf_id: str = "expected_wdf_id"
    wdf_values: list[str] = ["expected", "wdf", "values"]

    expected_setting: dict[str, Any] = {
        "data": {
            "attributes": {"filterValues": wdf_values},
            "id": wdf_setting_id,
            "relationships": {
                "workspaceDataFilter": {
                    "data": {"id": wdf_id, "type": "workspaceDataFilter"}
                }
            },
            "type": "workspaceDataFilterSetting",
        }
    }

    result_setting: dict[str, Any] = wdf_manager._create_wdf_setting_dict(
        wdf_setting_id, wdf_id, wdf_values
    )

    assert result_setting == expected_setting, (
        f"Expected {expected_setting}, but got {result_setting}"
    )


def test_get_wdf_settings_for_workspace_valid_payload(
    wdf_manager, mock_gooddata_api, mocker
) -> None:
    """Test processing of a valid response"""
    workspace_id: str = "expected_workspace_id"
    mock_response: Response = Response()
    mock_response.status_code = 200

    payload = WDF_VALID_PAYLOAD

    mock_response._content = orjson.dumps(payload)

    mocker.patch.object(
        mock_gooddata_api,
        "get_workspace_data_filter_settings",
        return_value=mock_response,
    )

    result = wdf_manager._get_wdf_settings_for_workspace(workspace_id)

    assert isinstance(result[0], WDFSetting), (
        f"Expected WDFSetting instance, but got {type(result)}"
    )


def test_get_wdf_settings_for_workspace_invalid_payload(
    wdf_manager, mock_gooddata_api, mocker
) -> None:
    """Test with an invalid payload -> will raise ValidationError / ValueError"""
    workspace_id: str = "expected_workspace_id"
    mock_response: Response = Response()
    mock_response.status_code = 200

    payload = {
        "data": [
            {
                "id": "expected_wdf_setting_id",
                "type": "workspaceDataFilterSetting",
                "attributes": {
                    "missing": "data",
                },
            }
        ]
    }

    mock_response._content = orjson.dumps(payload)

    mocker.patch.object(
        mock_gooddata_api,
        "get_workspace_data_filter_settings",
        return_value=mock_response,
    )

    with pytest.raises(ValidationError):
        wdf_manager._get_wdf_settings_for_workspace(workspace_id)


def test_get_actual_wdf_setting_id_and_values(wdf_manager) -> None:
    """Test getting the actual WDF setting ID and values."""
    data: dict[str, Any] = WDF_VALID_PAYLOAD["data"][0]
    actual_wdf_settings: list[WDFSetting] = [WDFSetting(**data)]
    wdf_id: str = "expected_wdf_id"

    actual_wdf_setting_id, actual_wdf_values = (
        wdf_manager._get_actual_wdf_setting_id_and_values(
            actual_wdf_settings, wdf_id
        )
    )

    assert actual_wdf_setting_id == "expected_wdf_setting_id", (
        f"Expected 'expected_wdf_setting_id', but got {actual_wdf_setting_id}"
    )

    assert actual_wdf_values == ["expected", "wdf", "values"], (
        f"Expected ['expected', 'wdf', 'values'], but got {actual_wdf_values}"
    )


def test_get_actual_wdf_setting_id_and_values_no_actuals(wdf_manager) -> None:
    """Should raise ValueError if no actuals are found"""
    actual_wdf_settings: list[WDFSetting] = []
    wdf_id: str = "expected_wdf_id"

    with pytest.raises(WorkspaceException):
        actual_wdf_setting_id, actual_wdf_values = (
            wdf_manager._get_actual_wdf_setting_id_and_values(
                actual_wdf_settings, wdf_id
            )
        )


def test_get_actual_wdf_setting_id_and_values_no_match(wdf_manager) -> None:
    """Should raise ValueError if no match is found"""
    data: dict[str, Any] = WDF_VALID_PAYLOAD["data"][0]
    actual_wdf_settings: list[WDFSetting] = [WDFSetting(**data)]
    wdf_id: str = "non_existent_wdf_id"

    with pytest.raises(WorkspaceException):
        actual_wdf_setting_id, actual_wdf_values = (
            wdf_manager._get_actual_wdf_setting_id_and_values(
                actual_wdf_settings, wdf_id
            )
        )


def test_compare_wdf_settings(wdf_manager, mocker) -> None:
    """Test the comparison of WDF settings."""
    workspace_context: WorkspaceContext = WorkspaceContext(
        workspace_id="workspace_id", workspace_name="workspace_name"
    )
    source_wdf_config: dict[str, list[str]] = {
        "expected_wdf_id": ["expected", "wdf", "values"],
        "expected_wdf_id_2": ["unexpected", "wdf", "values"],
        "expected_wdf_id_3": ["expected", "wdf", "values"],
    }
    actual_wdf_settings: list[WDFSetting] = [
        WDFSetting(**setting) for setting in WDF_ACTUAL_WDF_SETTINGS
    ]

    mock_put = mocker.patch.object(wdf_manager, "_put_wdf_setting")
    mock_post = mocker.patch.object(wdf_manager, "_post_wdf_setting")
    mock_delete = mocker.patch.object(
        wdf_manager, "_delete_redundant_wdf_setting"
    )

    wdf_manager._compare_wdf_settings(
        workspace_context,
        source_wdf_config,
        actual_wdf_settings,
    )

    assert mock_put.call_count == 1, "Expected one PUT call"
    assert mock_post.call_count == 1, "Expected one POST call"
    assert mock_delete.call_count == 1, "Expected one DELETE call"



# (C) 2025 GoodData Corporation

from typing import Any

import pytest

from gooddata_pipelines.utils.decorators import log_and_reraise_exception


@pytest.fixture()
def mocked_logger(mocker):
    return mocker.patch("gooddata_pipelines.utils.decorators.logger")


def test_log_and_re_raise_no_exception(mocked_logger):
    """Decorator should return inner function result and not log on success."""

    @log_and_reraise_exception("no-op")
    def target(a: int, b: int, *, c: int) -> int:
        return a + b + c

    result = target(1, 2, c=3)

    assert result == 6
    mocked_logger.error.assert_not_called()


def test_log_and_re_raise_logs_and_reraises(mocked_logger):
    """Decorator should log error and re-raise the original exception."""

    @log_and_reraise_exception("boom")
    def target(*args: Any, **kwargs: Any) -> None:
        raise ValueError("explosion")

    with pytest.raises(ValueError, match="explosion"):
        target(1, 2, a=3)

    mocked_logger.error.assert_called_once_with(
        "boom, target, Args: (1, 2), Kwargs: {'a': 3}"
    )



# (C) 2025 GoodData Corporation
"""This module defines type aliases intended to improve readability."""

from typing import TypeAlias

WorkspaceId: TypeAlias = str
DatasetId: TypeAlias = str

__all__ = ["WorkspaceId", "DatasetId"]



# (C) 2025 GoodData Corporation

"""Module for provisioning user permissions in GoodData workspaces."""

from typing import TypeVar

from gooddata_pipelines.api.exceptions import GoodDataApiException
from gooddata_pipelines.provisioning.entities.users.models.permissions import (
    EntityType,
    PermissionDeclaration,
    PermissionFullLoad,
    PermissionIncrementalLoad,
    TargetsPermissionDict,
    WSPermissionsDeclarations,
)
from gooddata_pipelines.provisioning.provisioning import Provisioning
from gooddata_pipelines.provisioning.utils.exceptions import BaseUserException

# Type variable for permission models (PermissionIncrementalLoad or PermissionFullLoad)
PermissionModel = TypeVar(
    "PermissionModel", PermissionIncrementalLoad, PermissionFullLoad
)


class PermissionProvisioner(
    Provisioning[PermissionFullLoad, PermissionIncrementalLoad]
):
    """Provisioning class for user permissions in GoodData workspaces.

    This class handles the provisioning of user permissions based on the provided
    source data. Use the `full_load` or `incremental_load`
    methods to run the provisioning.
    """

    source_group_incremental: list[PermissionIncrementalLoad]
    source_group_full: list[PermissionFullLoad]

    FULL_LOAD_TYPE: type[PermissionFullLoad] = PermissionFullLoad
    INCREMENTAL_LOAD_TYPE: type[PermissionIncrementalLoad] = (
        PermissionIncrementalLoad
    )

    def _get_ws_declaration(self, ws_id: str) -> PermissionDeclaration:
        users: TargetsPermissionDict = {}
        user_groups: TargetsPermissionDict = {}

        upstream_declaration = self._api.get_declarative_permissions(ws_id)

        for permission in upstream_declaration.permissions:
            permission_type, id = (
                permission.assignee.type,
                permission.assignee.id,
            )
            target_dict = (
                users
                if permission_type == EntityType.user.value
                else user_groups
            )

            id_permissions = target_dict.get(id)
            if not id_permissions:
                target_dict[id] = dict()

            target_dict[id][permission.name] = True

        return PermissionDeclaration(users, user_groups)

    def _get_upstream_declaration(
        self, ws_id: str
    ) -> PermissionDeclaration | None:
        """Retrieves upstream permission declaration for a workspace."""
        declaration = self._api.get_declarative_permissions(ws_id)
        return PermissionDeclaration.from_sdk_api(declaration)

    def _get_upstream_declarations(
        self, input_ws_ids: list[str]
    ) -> WSPermissionsDeclarations:
        """Retrieves upstream permission declarations for a list of workspaces."""
        ws_dict: WSPermissionsDeclarations = {}
        for ws_id in input_ws_ids:
            declaration = self._get_upstream_declaration(ws_id)
            if declaration:
                ws_dict[ws_id] = declaration
        return ws_dict

    @staticmethod
    def _construct_declarations(
        permissions: list[PermissionIncrementalLoad] | list[PermissionFullLoad],
    ) -> WSPermissionsDeclarations:
        """Constructs workspace permission declarations from the input permissions."""
        ws_dict: WSPermissionsDeclarations = {}
        for permission in permissions:
            ws_id = permission.workspace_id

            if ws_id not in ws_dict:
                ws_dict[ws_id] = PermissionDeclaration({}, {})

            if isinstance(permission, PermissionIncrementalLoad):
                ws_dict[ws_id].add_incremental_permission(permission)
            elif isinstance(permission, PermissionFullLoad):
                ws_dict[ws_id].add_full_load_permission(permission)
            else:
                raise ValueError(f"Invalid permission type: {type(permission)}")
        return ws_dict

    def _check_user_group_exists(self, ug_id: str) -> None:
        """Checks if user group with provided ID exists."""
        self._api._sdk.catalog_user.get_user_group(ug_id)

    def _validate_permission(
        self, permission: PermissionFullLoad | PermissionIncrementalLoad
    ) -> None:
        """Validates if the permission is correctly defined."""
        if permission.entity_type == EntityType.user:
            self._api.get_user(
                permission.entity_id, error_message="User not found"
            )
        else:
            self._api.get_user_group(
                permission.entity_id, error_message="User group not found"
            )

        self._api.get_workspace(
            permission.workspace_id, error_message="Workspace not found"
        )

    def _filter_invalid_permissions(
        self,
        permissions: list[PermissionModel],
    ) -> list[PermissionModel]:
        """Filters out invalid permissions from the input list."""
        valid_permissions: list[PermissionModel] = []

        for permission in permissions:
            try:
                self._validate_permission(permission)
            except (BaseUserException, GoodDataApiException) as e:
                self.logger.error(
                    f"Skipping {permission}. Error: {e.error_message} "
                    + f"Context: {permission.__dict__}"
                )
                continue
            valid_permissions.append(permission)
        return valid_permissions

    def _provision_incremental_load(self) -> None:
        """Provisiones permissions for a list of workspaces.

        Modifies existing upstream workspace permission declarations for each
        input workspace and skips rest of the workspaces.
        """
        valid_permissions = self._filter_invalid_permissions(
            self.source_group_incremental
        )

        input_declarations = self._construct_declarations(valid_permissions)

        input_ws_ids = list(input_declarations.keys())
        upstream_declarations = self._get_upstream_declarations(input_ws_ids)

        for ws_id, declaration in input_declarations.items():
            if ws_id not in upstream_declarations:
                continue

            upstream_declarations[ws_id].upsert(declaration)

            ws_permissions = upstream_declarations[ws_id].to_sdk_api()

            self._api.put_declarative_permissions(ws_id, ws_permissions)
            self.logger.info(f"Updated permissions for workspace {ws_id}")

    def _provision_full_load(self) -> None:
        """Provisions permissions for selected of workspaces.

        Modifies upstream workspace declarations for each input workspace and
        skips non-existent workspace ids. Overwrites any existing configuration
        of the workspace permissions.
        """
        valid_permissions = self._filter_invalid_permissions(
            self.source_group_full
        )

        input_declarations = self._construct_declarations(valid_permissions)

        for ws_id, declaration in input_declarations.items():
            ws_permissions = declaration.to_sdk_api()

            self._api.put_declarative_permissions(ws_id, ws_permissions)
            self.logger.info(f"Updated permissions for workspace {ws_id}")



# (C) 2025 GoodData Corporation

import pytest
from gooddata_sdk.catalog.workspace.entity_model.workspace import (
    CatalogWorkspace,
)

from gooddata_pipelines.provisioning import WorkspaceProvisioner
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceFullLoad,
)


@pytest.fixture
def test_find_workspaces_to_update_same_ids_and_names(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    ids_in_both_systems = {"workspace_id1", "workspace_id2"}
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="client_id1",
            workspace_id="workspace_id1",
            workspace_name="workspace_title1",
        ),
        WorkspaceFullLoad(
            parent_id="client_id2",
            workspace_id="workspace_id2",
            workspace_name="workspace_title2",
        ),
    ]
    panther_group: list[CatalogWorkspace] = [
        CatalogWorkspace(
            workspace_id="workspace_id1",
            name="workspace_title1",
            parent_id="parent_id",
        ),
        CatalogWorkspace(
            workspace_id="workspace_id2",
            name="workspace_title2",
            parent_id="parent_id",
        ),
    ]

    workspaces_to_update = workspace_provisioner._find_workspaces_to_update(
        source_group, panther_group, ids_in_both_systems
    )

    assert workspaces_to_update == set()


def test_find_workspaces_to_update_different_ids(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    ids_in_both_systems = {"workspace_id1", "workspace_id2"}
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="client_id1",
            workspace_id="workspace_id1",
            workspace_name="workspace_title1",
        ),
        WorkspaceFullLoad(
            parent_id="client_id2",
            workspace_id="workspace_id2",
            workspace_name="workspace_title2",
        ),
    ]
    panther_group: list[CatalogWorkspace] = [
        CatalogWorkspace(
            workspace_id="workspace_id1",
            name="workspace_title1",
            parent_id="parent_id",
        ),
        CatalogWorkspace(
            workspace_id="workspace_id2",
            name="workspace_title2",
            parent_id="parent_id",
        ),
    ]

    workspaces_to_update = workspace_provisioner._find_workspaces_to_update(
        source_group, panther_group, ids_in_both_systems
    )

    assert workspaces_to_update == set()


def test_find_workspaces_to_update_same_ids_different_names(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    ids_in_both_systems: set[str] = {"workspace_id1", "workspace_id2"}
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="client_id1",
            workspace_id="workspace_id1",
            workspace_name="workspace_title1",
        ),
        WorkspaceFullLoad(
            parent_id="client_id2",
            workspace_id="workspace_id2",
            workspace_name="workspace_title2",
        ),
    ]
    panther_group: list[CatalogWorkspace] = [
        CatalogWorkspace(
            workspace_id="workspace_id1",
            name="old_workspace_title1",
            parent_id="parent_id",
        ),
        CatalogWorkspace(
            workspace_id="workspace_id2",
            name="old_workspace_title2",
            parent_id="parent_id",
        ),
    ]

    workspaces_to_update = workspace_provisioner._find_workspaces_to_update(
        source_group, panther_group, ids_in_both_systems
    )

    assert workspaces_to_update == {"workspace_id1", "workspace_id2"}


def test_find_workspaces_to_update_no_panther(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    ids_in_both_systems: set[str] = set()
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="client_id1",
            workspace_id="workspace_id1",
            workspace_name="workspace_title1",
        ),
        WorkspaceFullLoad(
            parent_id="client_id2",
            workspace_id="workspace_id2",
            workspace_name="workspace_title2",
        ),
    ]
    panther_group: list[CatalogWorkspace] = []

    workspaces_to_update = workspace_provisioner._find_workspaces_to_update(
        source_group, panther_group, ids_in_both_systems
    )

    assert workspaces_to_update == set()



# (C) 2025 GoodData Corporation
"""Module containing models related to workspace provisioning in GoodData Cloud."""

from typing import Literal

import attrs
from pydantic import BaseModel, ConfigDict


@attrs.define
class WorkspaceDataMaps:
    """Dataclass to hold various mappings related to workspace data."""

    child_to_parent_id_map: dict[str, str] = attrs.field(factory=dict)
    workspace_id_to_wdf_map: dict[str, dict[str, list[str]]] = attrs.field(
        factory=dict
    )
    parent_ids: set[str] = attrs.field(factory=set)
    source_ids: set[str] = attrs.field(factory=set)
    workspace_id_to_name_map: dict[str, str] = attrs.field(factory=dict)
    upstream_ids: set[str] = attrs.field(factory=set)


class WorkspaceBase(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True, extra="forbid")

    parent_id: str
    workspace_id: str
    workspace_name: str
    workspace_data_filter_id: str | None = None
    workspace_data_filter_values: list[str] | None = None


class WorkspaceFullLoad(WorkspaceBase):
    """Input validator for full load of workspace provisioning."""


class WorkspaceIncrementalLoad(WorkspaceBase):
    """Input validator for incremental load of workspace provisioning."""

    is_active: bool


class WDFSettingAttributes(BaseModel):
    title: str
    filterValues: list[str]


class WDFSettingRelationshipsData(BaseModel):
    id: str
    type: Literal["workspaceDataFilter"]


class WDFSettingRelationships(BaseModel):
    workspaceDataFilter: dict[str, WDFSettingRelationshipsData]


class WDFSettingLinks(BaseModel):
    self: str


class WDFSettingMetaOrigin(BaseModel):
    originType: str
    originId: str


class WDFSettingMeta(BaseModel):
    origin: WDFSettingMetaOrigin


class WDFSetting(BaseModel):
    """Model representing a workspace data filter setting in GoodData Cloud."""

    id: str
    type: Literal["workspaceDataFilterSetting"]
    attributes: WDFSettingAttributes
    relationships: WDFSettingRelationships
    links: WDFSettingLinks
    meta: WDFSettingMeta



# (C) 2025 GoodData Corporation

"""Wrapper for interaction with GoodData Cloud."""

from gooddata_sdk.sdk import GoodDataSdk

from gooddata_pipelines.api.gooddata_api import ApiMethods
from gooddata_pipelines.api.gooddata_sdk import SdkMethods


class GoodDataApi(SdkMethods, ApiMethods):
    """Wrapper class for the GoodData Cloud API.

    This class combines interactions with the GoodData Python SDK and direct API
    calls.
    """

    def __init__(self, host: str, token: str) -> None:
        """Initialize the GoodDataApi with host and token.

        Args:
            host (str): The GoodData Cloud host URL.
            token (str): The authentication token for the GoodData Cloud API.
        """
        self._domain: str = host
        self._token: str = token

        # Initialize the GoodData SDK
        self._sdk = GoodDataSdk.create(self._domain, self._token)

        # Set up utils for direct API interaction
        self.base_url = self._get_base_url(self._domain)
        self.headers: dict = {
            "Authorization": f"Bearer {self._token}",
            "Content-Type": "application/vnd.gooddata.api+json",
        }



# (C) 2025 GoodData Corporation

"""Module for validating workspace data integrity in GoodData Cloud."""

from typing import Any

from requests import Response

from gooddata_pipelines.api import GoodDataApi
from gooddata_pipelines.logger.logger import LogObserver
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceFullLoad,
)
from gooddata_pipelines.provisioning.utils.context_objects import (
    WorkspaceContext,
)
from gooddata_pipelines.provisioning.utils.exceptions import (
    WorkspaceDataIntegrityException,
    WorkspaceException,
)


class WorkspaceDataValidator:
    """Class for validating workspace data integrity before provisioning."""

    def __init__(self, api: GoodDataApi):
        """
        Initializes the WorkspaceDataValidator with the GoodData API instance.

        Args:
            api (GoodDataApi): An instance of the GoodData API client.
        """
        self.api = api
        self.logger = LogObserver()

    def _check_basic_integrity(
        self,
        source_group: list[WorkspaceFullLoad],
    ) -> tuple[set[str], dict[str, list[str]]]:
        """
        Checks that mandatory fields are not empty and that that the combinations
        of values are unique.

        Returns a set of parent workspaces and a dictionary of parent-wdf mappings.
        """
        parent_workspaces: set[str] = set()
        parent_wdf_map: dict[str, list[str]] = {}
        parent_child_wdf_ids: list[tuple[str, str, str | None]] = []

        # Check that fields are not empty
        for workspace in source_group:
            parent_id: str | None = workspace.parent_id
            workspace_id: str | None = workspace.workspace_id
            workspace_name: str | None = workspace.workspace_name
            wdf_id: str | None = workspace.workspace_data_filter_id
            wdf_values: list[str] | None = (
                workspace.workspace_data_filter_values
            )

            # Create a context for the workspace validation
            validation_context: WorkspaceContext = WorkspaceContext(
                workspace_id=workspace_id,
                workspace_name=workspace_name,
                wdf_id=wdf_id,
                wdf_values=wdf_values,
            )

            # Raise specific error if both parent_id and workspace_id are not defined
            if (parent_id is None or parent_id == "") and (
                workspace_id is None or workspace_id == ""
            ):
                raise WorkspaceDataIntegrityException(
                    "Parent ID and workspace ID are not defined for at least one row. Please check the source data."
                )

            # Raise error if parent_id is not defined
            if parent_id is None or parent_id == "":
                raise WorkspaceDataIntegrityException(
                    "Parent ID is not defined in source data.",
                    validation_context,
                )

            # Add parent_id to the set of unique parent workspaces
            parent_workspaces.add(parent_id)

            # Raise error if workspace_id is not defined
            if workspace_id is None or workspace_id == "":
                raise WorkspaceDataIntegrityException(
                    f"Workspace ID is not defined for parent {parent_id}"
                )

            # Raise error if wdf_id is not defined but has values
            if wdf_id is not None and wdf_id != "":
                if wdf_values is None or wdf_values == []:
                    raise WorkspaceDataIntegrityException(
                        "WDF ID is defined but no WDF values are provided",
                        validation_context,
                    )

                # Add wdf_id to the parent-wdf dict if the value is defined
                if not parent_wdf_map.get(parent_id):
                    parent_wdf_map[parent_id] = []

                parent_wdf_map[parent_id].append(wdf_id)

            # Raise error if wdf_values are defined but wdf_id is not defined
            if wdf_values is not None and wdf_values != []:
                if wdf_id is None or wdf_id == "":
                    raise WorkspaceDataIntegrityException(
                        "WDF values are provided but WDF ID is not defined.",
                        validation_context,
                    )

            parent_child_wdf_ids.append((parent_id, workspace_id, wdf_id))

        # Check whether there are non-unique combinations in data
        if len(parent_child_wdf_ids) != len(set(parent_child_wdf_ids)):
            # Log the error to the database as a warning, but continue execution
            self.logger.warning(
                "Duplicate combinations of parent_id, workspace_id, "
                + "wdf_id exist in the source data."
            )

        return parent_workspaces, parent_wdf_map

    def _check_parent_exist(self, parent_id: str) -> None:
        """
        Raises an error if a parent workspace does not exist in Panther.
        """
        if not self.api.check_workspace_exists(parent_id):
            raise WorkspaceException(
                f"Parent workspace {parent_id} does not exist in Panther.",
                workspace_id=parent_id,
            )

    def _check_wdf_is_set_on_parent(
        self, parent_id: str, source_wdf_ids: list[str]
    ) -> None:
        """Raises an error if the parent workspace does not contain any of the defined wdf_id."""
        wdf_response: Response = self.api.get_all_workspace_data_filters(
            parent_id
        )
        wdf_json: dict[str, Any] = wdf_response.json()
        wdf_data: list[dict[str, Any]] = wdf_json.get("data", [])
        wdf_ids_on_parent: set[str] = {wdf["id"] for wdf in wdf_data}

        for source_wdf_id in source_wdf_ids:
            if source_wdf_id not in wdf_ids_on_parent:
                raise WorkspaceException(
                    f"WDF is not set on parent workspace {parent_id}.",
                    wdf_id=source_wdf_id,
                    workspace_id=parent_id,
                )

    def validate_source_data(
        self, source_group: list[WorkspaceFullLoad]
    ) -> None:
        """
        Validates the source data integrity.

        **Raises error when**:
        - the list of workspaces is empty
        - parent_id is not defined
        - workspace_id is not defined
        - The parent workspace does not exist
        - The parent workspace does not contain defined wdf_id.
        - wdf_id is defined but wdf_values is not defined
        - wdf_values are defined but wdf_id is not defined

        **Logs a warning when**:
        - There are more values for the parent_id, workspace_id, wdf_id combination.
        """
        if not source_group:
            # Raise error if source is empty
            raise WorkspaceException(
                "No workspaces found in the source database."
            )

        unique_parents, parent_wdf_map = self._check_basic_integrity(
            source_group
        )

        for parent_id in unique_parents:
            self._check_parent_exist(parent_id)

            self._check_wdf_is_set_on_parent(
                parent_id, parent_wdf_map[parent_id]
            )



# (C) 2025 GoodData Corporation
import orjson
import pytest

from gooddata_pipelines.ldm_extension.models.analytical_object import (
    AnalyticalObject,
    AnalyticalObjects,
)
from tests.conftest import TEST_DATA_DIR


@pytest.mark.parametrize(
    "file_path",
    [
        f"{TEST_DATA_DIR}/custom_fields/response_get_all_metrics.json",
        f"{TEST_DATA_DIR}/custom_fields/response_get_all_visualizations.json",
        f"{TEST_DATA_DIR}/custom_fields/response_get_all_dashboards.json",
    ],
)
def test_analytical_object_model_with_metrics(file_path):
    with open(file_path, "r") as file:
        data = orjson.loads(file.read())
    analytical_objects = AnalyticalObjects(**data)
    assert isinstance(analytical_objects, AnalyticalObjects)
    assert isinstance(analytical_objects.data, list)
    assert all(
        isinstance(obj, AnalyticalObject) for obj in analytical_objects.data
    )


@pytest.mark.parametrize(
    "response",
    [
        {
            "something": "unexpected",
        },
        {
            "data": [
                {
                    # "id": "metric1", # Missing id field
                    "type": "metric",
                    "attributes": {
                        "title": "Test Metric",
                        "areRelationsValid": True,
                    },
                }
            ]
        },
        {
            "data": [
                {
                    "id": 123,  # invalid id type
                    "type": "metric",
                    "attributes": {
                        "title": "Test Metric",
                        "areRelationsValid": True,
                    },
                }
            ]
        },
    ],
)
def test_analytical_object_model_with_invalid_response(response):
    with pytest.raises(ValueError):
        AnalyticalObjects(**response)



# (C) 2025 GoodData Corporation
"""This module defines enums and models used to represent the input data.

Models defined here are used to validate and structure the input data before
further processing.
"""

from enum import Enum

from pydantic import BaseModel, model_validator


class CustomFieldType(str, Enum):
    """GoodData field types."""

    # NOTE: Start using StrEnum with Python 3.11
    ATTRIBUTE = "attribute"
    FACT = "fact"
    DATE = "date"


class ColumnDataType(str, Enum):
    """Supported data types"""

    # NOTE: Start using StrEnum with Python 3.11
    INT = "INT"
    STRING = "STRING"
    DATE = "DATE"
    NUMERIC = "NUMERIC"
    TIMESTAMP = "TIMESTAMP"
    TIMESTAMP_TZ = "TIMESTAMP_TZ"
    BOOLEAN = "BOOLEAN"


class CustomFieldDefinition(BaseModel):
    """Input model for custom field definition."""

    workspace_id: str
    dataset_id: str
    custom_field_id: str
    custom_field_name: str
    custom_field_type: CustomFieldType
    custom_field_source_column: str
    custom_field_source_column_data_type: ColumnDataType

    @model_validator(mode="after")
    def check_ids_not_equal(self) -> "CustomFieldDefinition":
        """Check that custom field ID is not the same as dataset ID."""
        if self.custom_field_id == self.dataset_id:
            raise ValueError(
                f"Custom field ID {self.custom_field_id} cannot be the same as dataset ID {self.dataset_id}"
            )
        return self


class CustomDatasetDefinition(BaseModel):
    """Input model for custom dataset definition."""

    workspace_id: str
    dataset_id: str
    dataset_name: str
    dataset_datasource_id: str
    dataset_source_table: str | None
    dataset_source_sql: str | None
    parent_dataset_reference: str
    parent_dataset_reference_attribute_id: str
    dataset_reference_source_column: str
    dataset_reference_source_column_data_type: ColumnDataType
    workspace_data_filter_id: str
    workspace_data_filter_column_name: str

    @model_validator(mode="after")
    def check_source(self) -> "CustomDatasetDefinition":
        """At least one of dataset_source_table or dataset_source_sql is provided."""
        if not (self.dataset_source_table or self.dataset_source_sql):
            raise ValueError(
                "One of dataset_source_table and dataset_source_sql must be provided"
            )
        if self.dataset_source_table and self.dataset_source_sql:
            raise ValueError(
                "Only one of dataset_source_table and dataset_source_sql can be provided"
            )
        return self


class CustomDataset(BaseModel):
    """Custom dataset with its definition and custom fields."""

    definition: CustomDatasetDefinition
    custom_fields: list[CustomFieldDefinition]



# (C) 2025 GoodData Corporation

import os
import shutil
import tempfile
from pathlib import Path
from unittest import mock

import boto3
import pytest
from moto import mock_aws

from gooddata_pipelines.backup_and_restore.backup_manager import (
    BackupBatch,
    BackupManager,
)
from gooddata_pipelines.backup_and_restore.constants import BackupSettings
from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
    LocalStorageConfig,
    S3StorageConfig,
    StorageType,
)
from gooddata_pipelines.backup_and_restore.storage.local_storage import (
    LocalStorage,
)
from gooddata_pipelines.backup_and_restore.storage.s3_storage import S3Storage
from tests.conftest import TEST_DATA_DIR

TEST_DATA_SUBDIR = f"{TEST_DATA_DIR}/backup"

S3_BACKUP_PATH = "some/s3/backup/path/org_id/"
S3_BUCKET = "some-s3-bucket"

LOCAL_CONFIG = BackupRestoreConfig(
    storage_type=StorageType.LOCAL,
    storage=LocalStorageConfig(backup_path=f"{TEST_DATA_DIR}/local_export"),
)

S3_CONFIG = BackupRestoreConfig(
    storage_type=StorageType.S3,
    storage=S3StorageConfig(
        bucket=S3_BUCKET,
        backup_path=S3_BACKUP_PATH,
        profile="default",
    ),
)


@pytest.fixture
def backup_manager(mock_logger):
    with (
        mock.patch.object(BackupManager, "_api", create=True),
        mock.patch(
            "gooddata_pipelines.api.gooddata_api_wrapper.GoodDataApi.get_organization_id",
            return_value="services",
        ),
    ):
        manager = BackupManager.create(
            S3_CONFIG,
            "host",
            "token",
        )
        manager.logger.subscribe(mock_logger)
        return manager


@pytest.fixture()
def s3():
    with mock_aws():
        yield boto3.resource("s3")


@pytest.fixture(scope="function")
def s3_bucket(s3):
    s3.create_bucket(Bucket=S3_BUCKET)
    yield s3.Bucket(S3_BUCKET)


@pytest.fixture(scope="function")
def create_backups_in_bucket(s3_bucket):
    def create_backups(
        ws_ids: list[str], is_e2e: bool = False, suffix: str = "bla"
    ):
        # If used within e2e test, add some suffix to path
        # in order to simulate a more realistic scenario
        path_suffix = f"/{suffix}" if is_e2e else ""

        for ws_id in ws_ids:
            s3_bucket.put_object(
                Bucket=S3_BUCKET, Key=f"{S3_BACKUP_PATH}{ws_id}{path_suffix}/"
            )
            s3_bucket.put_object(
                Bucket=S3_BUCKET,
                Key=f"{S3_BACKUP_PATH}{ws_id}{path_suffix}/gooddata_layouts.zip",
            )

    return create_backups


def assert_not_called_with(target, *args, **kwargs):
    try:
        target.assert_called_with(*args, **kwargs)
    except AssertionError:
        return
    formatted_call = target._format_mock_call_signature(args, kwargs)
    raise AssertionError(f"Expected {formatted_call} to not have been called.")


def test_get_s3_storage(backup_manager):
    """Test get_storage method with literal string as input."""
    s3_storage = backup_manager._get_storage(S3_CONFIG)
    assert isinstance(s3_storage, S3Storage)


def test_get_local_storage(backup_manager):
    """Test get_storage method with literal string as input."""
    local_storage = backup_manager._get_storage(LOCAL_CONFIG)
    assert isinstance(local_storage, LocalStorage)


# Test that zipping gooddata_layouts folder works
def test_archive_gooddata_layouts_to_zip(backup_manager):
    with tempfile.TemporaryDirectory() as tmpdir:
        shutil.copytree(
            Path(
                f"{TEST_DATA_SUBDIR}/test_exports/services/",
            ),
            Path(tmpdir + "/services"),
        )
        backup_manager._archive_gooddata_layouts_to_zip(
            str(Path(tmpdir, "services"))
        )

        zip_exists = os.path.isfile(
            Path(
                tmpdir,
                "services/wsid1/20230713-132759-1_3_1_dev5/gooddata_layouts.zip",
            )
        )
        gooddata_layouts_dir_exists = os.path.isdir(
            Path(
                tmpdir,
                "services/wsid1/20230713-132759-1_3_1_dev5/gooddata_layouts",
            )
        )

        assert gooddata_layouts_dir_exists is False
        assert zip_exists

        zip_exists = os.path.isfile(
            Path(
                tmpdir,
                "services/wsid2/20230713-132759-1_3_1_dev5/gooddata_layouts.zip",
            )
        )
        gooddata_layouts_dir_exists = os.path.isdir(
            Path(
                tmpdir,
                "services/wsid2/20230713-132759-1_3_1_dev5/gooddata_layouts",
            )
        )

        assert gooddata_layouts_dir_exists is False
        assert zip_exists

        zip_exists = os.path.isfile(
            Path(
                tmpdir,
                "services/wsid3/20230713-132759-1_3_1_dev5/gooddata_layouts.zip",
            )
        )
        gooddata_layouts_dir_exists = os.path.isdir(
            Path(
                tmpdir,
                "services/wsid3/20230713-132759-1_3_1_dev5/gooddata_layouts",
            )
        )

        assert gooddata_layouts_dir_exists is False
        assert zip_exists


def test_store_user_data_filters(backup_manager):
    user_data_filters = {
        "userDataFilters": [
            {
                "id": "datafilter2",
                "maql": '{label/campaign_channels.category} = "1"',
                "title": "Status filter",
                "user": {
                    "id": "5c867a8a-12af-45bf-8d85-c7d16bedebd1",
                    "type": "user",
                },
            },
            {
                "id": "datafilter4",
                "maql": '{label/campaign_channels.category} = "1"',
                "title": "Status filter",
                "user": {
                    "id": "5c867a8a-12af-45bf-8d85-c7d16bedebd1",
                    "type": "user",
                },
            },
        ]
    }
    user_data_filter_folderlocation = f"{TEST_DATA_SUBDIR}/test_exports/services/wsid1/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid1/user_data_filters"
    backup_manager._store_user_data_filters(
        user_data_filters,
        Path(
            f"{TEST_DATA_SUBDIR}/test_exports/services/wsid1/20230713-132759-1_3_1_dev5",
        ),
        "wsid1",
    )
    user_data_filter_folder = os.path.isdir(
        Path(user_data_filter_folderlocation)
    )
    user_data_filter2 = os.path.isfile(
        Path(f"{user_data_filter_folderlocation}/datafilter2.yaml")
    )
    user_data_filter4 = os.path.isfile(
        Path(f"{user_data_filter_folderlocation}/datafilter4.yaml")
    )
    assert user_data_filter_folder
    assert user_data_filter2
    assert user_data_filter4

    count = 0
    for path in os.listdir(user_data_filter_folderlocation):
        if os.path.isfile(os.path.join(user_data_filter_folderlocation, path)):
            count += 1

    assert count == 2

    shutil.rmtree(
        Path(
            f"{TEST_DATA_SUBDIR}/test_exports/services/wsid1/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid1/user_data_filters",
        )
    )


def test_local_storage_export(backup_manager):
    with tempfile.TemporaryDirectory() as tmpdir:
        org_store_location = Path(tmpdir + "/services")
        shutil.copytree(
            Path(
                f"{TEST_DATA_SUBDIR}/test_exports/services/",
            ),
            org_store_location,
        )
        local_storage = backup_manager._get_storage(LOCAL_CONFIG)

        local_storage.export(
            folder=tmpdir,
            org_id="services",
        )

        local_export_folder_exist = os.path.isdir(
            Path(
                f"{TEST_DATA_DIR}/local_export/services/wsid1/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid1/analytics_model"
            )
        )
        local_export_folder2_exist = os.path.isdir(
            Path(
                f"{TEST_DATA_DIR}/local_export/services/wsid3/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid3/ldm"
            )
        )

        local_export_folder3_exist = os.path.isdir(
            Path(
                f"{TEST_DATA_DIR}/local_export/services/wsid3/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid3/user_data_filters"
            )
        )

        local_export_file_exist = os.path.isfile(
            Path(
                f"{TEST_DATA_DIR}/local_export/services/wsid2/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid2/analytics_model/analytical_dashboards/id.yaml"
            )
        )
        assert local_export_folder_exist
        assert local_export_folder2_exist
        assert local_export_folder3_exist
        assert local_export_file_exist
        shutil.rmtree(f"{TEST_DATA_DIR}/local_export")


def test_file_upload(backup_manager, s3, s3_bucket):
    backup_manager.storage.export(
        f"{TEST_DATA_SUBDIR}/test_exports", "services"
    )
    s3.Object(
        S3_BUCKET,
        "some/s3/backup/path/org_id/services/wsid2/20230713-132759-1_3_1_dev5/gooddata_layouts/services/workspaces/wsid2/analytics_model/filter_contexts/id.yaml",
    ).load()


def test_split_to_batches(backup_manager):
    workspaces = ["ws1", "ws2", "ws3", "ws4", "ws5"]
    batch_size = 2
    expected_batches = [
        BackupBatch(["ws1", "ws2"]),
        BackupBatch(["ws3", "ws4"]),
        BackupBatch(["ws5"]),
    ]

    result = backup_manager._split_to_batches(workspaces, batch_size)

    for i, batch in enumerate(result):
        assert isinstance(batch, BackupBatch)
        assert batch.list_of_ids == expected_batches[i].list_of_ids


@mock.patch(
    "gooddata_pipelines.backup_and_restore.backup_manager.BackupManager._get_workspace_export"
)
@mock.patch(
    "gooddata_pipelines.backup_and_restore.backup_manager.BackupManager._archive_gooddata_layouts_to_zip"
)
def test_process_batch_success(
    archive_gooddata_layouts_to_zip_mock,
    get_workspace_export_mock,
    backup_manager,
):
    # Mock the storage's export method
    backup_manager.storage = mock.Mock()
    batch = BackupBatch(["ws1", "ws2"])

    backup_manager._process_batch(
        batch=batch,
        retry_count=0,
    )

    get_workspace_export_mock.assert_called_once()
    archive_gooddata_layouts_to_zip_mock.assert_called_once()
    backup_manager.storage.export.assert_called_once()


@mock.patch(
    "gooddata_pipelines.backup_and_restore.backup_manager.BackupManager._get_workspace_export"
)
@mock.patch(
    "gooddata_pipelines.backup_and_restore.backup_manager.BackupManager._archive_gooddata_layouts_to_zip"
)
def test_process_batch_retries_on_exception(
    archive_gooddata_layouts_to_zip_mock,
    get_workspace_export_mock,
    backup_manager,
    capsys,
):
    backup_manager.storage = mock.Mock()
    batch = BackupBatch(["ws1"])

    # Raise exception on first call, succeed on second
    call_count = {"count": 0}

    def fail_once(*args, **kwargs):
        if call_count["count"] == 0:
            call_count["count"] += 1
            raise Exception("fail")
        return None

    get_workspace_export_mock.side_effect = fail_once

    backup_manager._process_batch(
        batch=batch,
    )

    assert get_workspace_export_mock.call_count == 2
    captured = capsys.readouterr()
    assert captured.out.startswith(
        "Exception encountered while processing a batch. Retrying"
    )
    backup_manager.storage.export.assert_called_once()


@mock.patch(
    "gooddata_pipelines.backup_and_restore.backup_manager.BackupManager._get_workspace_export"
)
@mock.patch(
    "gooddata_pipelines.backup_and_restore.backup_manager.BackupManager._archive_gooddata_layouts_to_zip"
)
def test_process_batch_raises_after_max_retries(
    archive_gooddata_layouts_to_zip_mock,
    get_workspace_export_mock,
    backup_manager,
    capsys,
):
    backup_manager.storage = mock.Mock()
    batch = BackupBatch(["ws1"])
    get_workspace_export_mock.side_effect = Exception("fail")

    with pytest.raises(Exception) as exc_info:
        backup_manager._process_batch(
            batch=batch,
            retry_count=BackupSettings.MAX_RETRIES,
        )
    assert str(exc_info.value) == "fail"
    captured = capsys.readouterr()
    assert captured.out.startswith("Batch failed:")



# (C) 2025 GoodData Corporation

"""Provisioning base class for GoodData Pipelines."""

from pathlib import Path
from typing import Generic, Type, TypeVar

from gooddata_sdk.utils import PROFILES_FILE_PATH, profile_content

from gooddata_pipelines.api import GoodDataApi
from gooddata_pipelines.logger.logger import (
    LogObserver,
)
from gooddata_pipelines.provisioning.utils.utils import EntityGroupIds

TFullLoadSourceData = TypeVar("TFullLoadSourceData")
TIncrementalSourceData = TypeVar("TIncrementalSourceData")


class Provisioning(Generic[TFullLoadSourceData, TIncrementalSourceData]):
    """Base provisioning class."""

    TProvisioning = TypeVar("TProvisioning", bound="Provisioning")
    source_group_full: list[TFullLoadSourceData]
    source_group_incremental: list[TIncrementalSourceData]

    FULL_LOAD_TYPE: type[TFullLoadSourceData]
    INCREMENTAL_LOAD_TYPE: type[TIncrementalSourceData]

    def __init__(self, host: str, token: str) -> None:
        self.source_id: set[str] = set()
        self.upstream_id: set[str] = set()
        self._api = GoodDataApi(host, token)
        self.logger: LogObserver = LogObserver()
        self.fatal_exception: str = ""

    @classmethod
    def create(
        cls: Type[TProvisioning], host: str, token: str
    ) -> TProvisioning:
        """Creates a provisioner instance using provided host and token."""
        cls._validate_credentials(host, token)
        return cls(host=host, token=token)

    @classmethod
    def create_from_profile(
        cls: Type[TProvisioning],
        profile: str = "default",
        profiles_path: Path = PROFILES_FILE_PATH,
    ) -> TProvisioning:
        """Creates a provisioner instance using a GoodData profile file."""
        content = profile_content(profile, profiles_path)
        return cls(host=content["host"], token=content["token"])

    @staticmethod
    def _validate_credentials(host: str, token: str) -> None:
        """Validates the credentials."""
        if (not host) and (not token):
            raise ValueError("Host and token are required.")
        if not host:
            raise ValueError("Host is required.")
        if not token:
            raise ValueError("Token is required.")

    @staticmethod
    def _create_groups(
        source_id: set[str], panther_id: set[str]
    ) -> EntityGroupIds:
        """Creates groups for provisioning as sets of IDs.

        Sorts the IDs into three categories:
        - IDs that exist both source and upstream (to be checked further)
        - IDs that exist upstream but not in source (to be deleted)
        - IDs that exist in source but not upstream (to be created)
        """
        ids_in_both_systems: set[str] = source_id.intersection(panther_id)
        ids_to_delete: set[str] = panther_id.difference(source_id)
        ids_to_create: set[str] = source_id.difference(panther_id)

        return EntityGroupIds(
            ids_in_both_systems=ids_in_both_systems,
            ids_to_delete=ids_to_delete,
            ids_to_create=ids_to_create,
        )

    def _validate_source_data_type(
        self,
        source_data: list[TFullLoadSourceData] | list[TIncrementalSourceData],
        model: type[TFullLoadSourceData] | type[TIncrementalSourceData],
    ) -> None:
        """Validates data type of the source data."""
        if not all(isinstance(record, model) for record in source_data):
            raise TypeError(
                f"Not all elements in source data are instances of {model.__name__}"
            )

    def _provision_incremental_load(self) -> None:
        raise NotImplementedError(
            "Provisioning method to be implemented in the subclass."
        )

    def _provision_full_load(self) -> None:
        raise NotImplementedError(
            "Provisioning method to be implemented in the subclass."
        )

    def full_load(self, source_data: list[TFullLoadSourceData]) -> None:
        """Runs full provisioning workflow with the provided source data.

        Full provisioning is a full load of the source data, where the source data
        is assumed to a single source of truth and the upstream workspaces are updated
        to match it.

        That means:
        - All workspaces declared in the source data are created if missing, or
        updated to match the source data
        - All child workspaces not declared under the parent workspace in the
        source data are deleted
        """

        try:
            self._validate_source_data_type(source_data, self.FULL_LOAD_TYPE)
            self.source_group_full = source_data
            self._provision_full_load()
            self.logger.info("Provisioning completed.")
        except Exception as e:
            self._handle_fatal_exception(e)

    def incremental_load(
        self, source_data: list[TIncrementalSourceData]
    ) -> None:
        """Runs incremental provisioning workflow with the provided source data.

        Incremental provisioning is used to modify a subset of the upstream workspaces
        based on the source data provided. Only changes requested in the source
        data will be applied.
        """
        try:
            self._validate_source_data_type(
                source_data, self.INCREMENTAL_LOAD_TYPE
            )
            self.source_group_incremental = source_data
            self._provision_incremental_load()
            self.logger.info("Provisioning completed.")
        except Exception as e:
            self._handle_fatal_exception(e)

    def _handle_fatal_exception(self, e: Exception) -> None:
        """Handles fatal exceptions during provisioning.

        Logs the exception content. Re-raises the exception if there is no
        subscriber to the logger.
        """
        self.fatal_exception = str(e)

        if hasattr(e, "__dict__"):
            exception_context = f"Context: {e.__dict__}"
        else:
            exception_context = ""

        exception_message = (
            f"Provisioning failed. Error: {self.fatal_exception}. "
            + exception_context
        )

        self.logger.error(exception_message)

        raise Exception(exception_message)



# (C) 2025 GoodData Corporation

from typing import Any

WDF_VALID_PAYLOAD = {
    "data": [
        {
            "id": "expected_wdf_setting_id",
            "type": "workspaceDataFilterSetting",
            "attributes": {
                "title": "",
                "filterValues": ["expected", "wdf", "values"],
            },
            "relationships": {
                "workspaceDataFilter": {
                    "data": {
                        "id": "expected_wdf_id",
                        "type": "workspaceDataFilter",
                    },
                },
            },
            "links": {"self": "https://some.uri.com"},
            "meta": {
                "origin": {
                    "originType": "origin_type",
                    "originId": "origin_id",
                }
            },
        }
    ]
}

WDF_ACTUAL_WDF_SETTINGS: list[dict[str, Any]] = [
    {
        "id": "expected_wdf_setting_id",
        "type": "workspaceDataFilterSetting",
        "attributes": {
            "title": "",
            "filterValues": ["expected", "wdf", "values"],
        },
        "relationships": {
            "workspaceDataFilter": {
                "data": {
                    "id": "expected_wdf_id",
                    "type": "workspaceDataFilter",
                },
            },
        },
        "links": {"self": "https://some.uri.com"},
        "meta": {
            "origin": {
                "originType": "origin_type",
                "originId": "workspace_id",
            }
        },
    },
    {
        "id": "expected_wdf_setting_id_2",
        "type": "workspaceDataFilterSetting",
        "attributes": {
            "title": "",
            "filterValues": ["expected", "wdf", "values"],
        },
        "relationships": {
            "workspaceDataFilter": {
                "data": {
                    "id": "expected_wdf_id_2",
                    "type": "workspaceDataFilter",
                },
            },
        },
        "links": {"self": "https://some.uri.com"},
        "meta": {
            "origin": {
                "originType": "origin_type",
                "originId": "workspace_id",
            }
        },
    },
    # expected_wdf_id_3 is missing
    {
        "id": "expected_wdf_setting_id_4",
        "type": "workspaceDataFilterSetting",
        "attributes": {
            "title": "",
            "filterValues": ["expected", "wdf", "values"],
        },
        "relationships": {
            "workspaceDataFilter": {
                "data": {
                    "id": "expected_wdf_id_4",
                    "type": "workspaceDataFilter",
                },
            },
        },
        "links": {"self": "https://some.uri.com"},
        "meta": {
            "origin": {
                "originType": "origin_type",
                "originId": "workspace_id",
            }
        },
    },
]



# (C) 2025 GoodData Corporation
import pytest

from gooddata_pipelines.ldm_extension.input_validator import (
    LdmExtensionDataValidator,
)
from gooddata_pipelines.ldm_extension.models.custom_data_object import (
    ColumnDataType,
    CustomDataset,
    CustomDatasetDefinition,
    CustomFieldDefinition,
    CustomFieldType,
)


@pytest.fixture
def valid_dataset_definitions():
    """Fixture to provide valid dataset definitions for testing."""
    return [
        CustomDatasetDefinition(
            workspace_id="ws1",
            dataset_id="ds1",
            dataset_name="Dataset 1",
            dataset_datasource_id="ds_source_1",
            dataset_source_table="table1",
            dataset_source_sql=None,
            parent_dataset_reference="parent1",
            parent_dataset_reference_attribute_id="parent1.id",
            dataset_reference_source_column="id",
            dataset_reference_source_column_data_type=ColumnDataType.STRING,
            workspace_data_filter_id="wdf1",
            workspace_data_filter_column_name="id",
        ),
        CustomDatasetDefinition(
            workspace_id="ws2",
            dataset_id="ds1",
            dataset_name="Dataset 2",
            dataset_datasource_id="ds_source_2",
            dataset_source_table="table2",
            dataset_source_sql=None,
            parent_dataset_reference="parent2",
            parent_dataset_reference_attribute_id="parent2.id",
            dataset_reference_source_column="id",
            dataset_reference_source_column_data_type=ColumnDataType.INT,
            workspace_data_filter_id="wdf2",
            workspace_data_filter_column_name="id",
        ),
    ]


@pytest.fixture
def valid_field_definitions():
    """Fixture to provide valid field definitions for testing."""
    return [
        CustomFieldDefinition(
            workspace_id="ws1",
            dataset_id="ds1",
            custom_field_id="cf1",
            custom_field_name="Field 1",
            custom_field_type=CustomFieldType.ATTRIBUTE,
            custom_field_source_column="col1",
            custom_field_source_column_data_type=ColumnDataType.STRING,
        ),
        CustomFieldDefinition(
            workspace_id="ws1",
            dataset_id="ds1",
            custom_field_id="cf2",
            custom_field_name="Field 2",
            custom_field_type=CustomFieldType.ATTRIBUTE,
            custom_field_source_column="col2",
            custom_field_source_column_data_type=ColumnDataType.STRING,
        ),
        CustomFieldDefinition(
            workspace_id="ws2",
            dataset_id="ds1",
            custom_field_id="cf3",
            custom_field_name="Field 3",
            custom_field_type=CustomFieldType.ATTRIBUTE,
            custom_field_source_column="col3",
            custom_field_source_column_data_type=ColumnDataType.STRING,
        ),
    ]


def test_validate_success(valid_dataset_definitions, valid_field_definitions):
    """Provide valid input data and expect successful validation."""
    validator = LdmExtensionDataValidator()
    result = validator.validate(
        valid_dataset_definitions, valid_field_definitions
    )
    assert isinstance(result, dict)
    assert "ws1" in result
    assert "ds1" in result["ws1"]
    assert isinstance(result["ws1"]["ds1"], CustomDataset)
    assert len(result["ws1"]["ds1"].custom_fields) == 2
    assert result["ws2"]["ds1"].custom_fields[0].custom_field_id == "cf3"


def test_duplicate_dataset_raises(valid_dataset_definitions):
    """Test that duplicate dataset definitions raise a ValueError."""
    # Add a duplicate dataset definition
    invalid = valid_dataset_definitions + valid_dataset_definitions
    validator = LdmExtensionDataValidator()
    with pytest.raises(ValueError, match="Duplicate dataset definitions"):
        validator.validate(invalid, [])


def test_duplicate_field_workspace_level(valid_dataset_definitions):
    """Duplicate cf_id for ATTRIBUTE in same workspace. should raise ValueError."""
    fields = [
        CustomFieldDefinition(
            workspace_id="ws1",
            dataset_id="ds1",
            custom_field_id="cf1",
            custom_field_type=CustomFieldType.ATTRIBUTE,
            custom_field_name="Field 1",
            custom_field_source_column="col1",
            custom_field_source_column_data_type=ColumnDataType.STRING,
        ),
        CustomFieldDefinition(
            workspace_id="ws1",
            dataset_id="ds2",
            custom_field_id="cf1",
            custom_field_type=CustomFieldType.ATTRIBUTE,
            custom_field_name="Field 2",
            custom_field_source_column="col2",
            custom_field_source_column_data_type=ColumnDataType.STRING,
        ),
    ]
    validator = LdmExtensionDataValidator()
    with pytest.raises(
        ValueError,
        match="Duplicate custom field found for workspace ws1 with field ID cf1",
    ):
        validator.validate(valid_dataset_definitions, fields)


def test_duplicate_field_dataset_level(valid_dataset_definitions):
    """Duplicate cf_id for DATE in same dataset. should raise ValueError."""
    fields = [
        CustomFieldDefinition(
            workspace_id="ws1",
            dataset_id="ds1",
            custom_field_id="cf1",
            custom_field_type=CustomFieldType.DATE,
            custom_field_name="Field 1",
            custom_field_source_column="col1",
            custom_field_source_column_data_type=ColumnDataType.DATE,
        ),
        CustomFieldDefinition(
            workspace_id="ws1",
            dataset_id="ds1",
            custom_field_id="cf1",
            custom_field_type=CustomFieldType.DATE,
            custom_field_name="Field 2",
            custom_field_source_column="col2",
            custom_field_source_column_data_type=ColumnDataType.DATE,
        ),
    ]
    validator = LdmExtensionDataValidator()
    with pytest.raises(
        ValueError,
        match="Duplicate custom field found for dataset ds1 with field ID cf1",
    ):
        validator.validate(valid_dataset_definitions, fields)



# (C) 2025 GoodData Corporation
"""This module defines the AnalyticalObjects Pydantic model.

The model is used to represent features of analytical objects important for
checking the validity of references.
"""

from pydantic import BaseModel, Field


class Attributes(BaseModel):
    title: str
    are_relations_valid: bool = Field(alias="areRelationsValid")


class AnalyticalObject(BaseModel):
    id: str
    type: str
    attributes: Attributes


class AnalyticalObjects(BaseModel):
    """Simplified model representing response obtained from GoodData API when querying
    analytical objects.

    This model is used to represent analytical objects such as metrics, visualizations,
    and dashboard in a simplified manner, with the purpose of checkinf the validity
    of references of these objects.

    This is not a complete schema of the analytical objects!
    """

    data: list[AnalyticalObject]



# (C) 2025 GoodData Corporation


import pytest

from gooddata_pipelines.provisioning.entities.users.models.user_groups import (
    UserGroupIncrementalLoad,
)


def test_missing_key_no_parent_groups() -> None:
    result = UserGroupIncrementalLoad(
        user_group_id="ug_2", user_group_name="Developers", is_active=True
    )
    expected = UserGroupIncrementalLoad(
        user_group_id="ug_2",
        user_group_name="Developers",
        parent_user_groups=[],
        is_active=True,
    )

    assert result == expected, (
        "Row without parent user groups should be parsed correctly"
    )


@pytest.mark.parametrize("invalid_name", [None, ""])
def test_fallback_name(invalid_name) -> None:
    result = UserGroupIncrementalLoad(
        user_group_id="ug_3",
        user_group_name=invalid_name,
        parent_user_groups=[],
        is_active=False,
    )
    expected = UserGroupIncrementalLoad(
        user_group_id="ug_3",
        user_group_name="ug_3",
        parent_user_groups=[],
        is_active=False,
    )

    assert result == expected, (
        "Row with empty name should fallback to user group ID"
    )


@pytest.mark.parametrize("empty_parent_groups", [None, "", []])
def test_no_parent_user_groups(empty_parent_groups) -> None:
    result = UserGroupIncrementalLoad(
        user_group_id="ug_3",
        user_group_name="ug_3",
        parent_user_groups=empty_parent_groups,
        is_active=False,
    )
    expected = UserGroupIncrementalLoad(
        user_group_id="ug_3",
        user_group_name="ug_3",
        parent_user_groups=[],
        is_active=False,
    )

    assert result == expected, (
        "Row with empty parent user groups should be parsed correctly"
    )


def test_row_invalid_is_active() -> None:
    with pytest.raises(ValueError):
        UserGroupIncrementalLoad(
            user_group_id="ug_4",
            user_group_name="Testers",
            parent_user_groups=[],
            is_active="not_a_boolean",  # type: ignore
        )



# (C) 2025 GoodData Corporation

import pytest

from gooddata_pipelines.api.gooddata_api import (
    API_VERSION,
    ApiMethods,
)


@pytest.mark.parametrize(
    "domain, expected_base_url",
    [
        ("example.com", f"https://example.com/api/{API_VERSION}"),
        (
            "https://example.com",
            f"https://example.com/api/{API_VERSION}",
        ),
        (
            "http://example.com",
            f"https://example.com/api/{API_VERSION}",
        ),
        ("example.com/", f"https://example.com/api/{API_VERSION}"),
        (
            "https://example.com/",
            f"https://example.com/api/{API_VERSION}",
        ),
        (
            "http://example.com/",
            f"https://example.com/api/{API_VERSION}",
        ),
    ],
)
def test_get_base_url(domain, expected_base_url):
    """Test the get_base_url method with various domain inputs."""
    result = ApiMethods._get_base_url(domain)
    assert result == expected_base_url
